# A Study of Learning Search Approximation in Mixed Integer Branch and Bound: Node Selection in SCIP

## Prerequisites
### Technical
- Install Conda
- Use Conda environment `environment.yml` in the project root.

### Shared variables
- Choose a problem: `{setcover, cauctions, facilities, indset}`
- Choose a difficulty (problem dependent), formats:
    - `setcover: {x}r_{y}c_{z}d (x = rows, y = columns, d = density, e.g. 1000r_2000c_0.05d)`
    - `cauctions: {x}_{y} (x = number of items, y = number of bids, e.g. 250_1200)`
    - `facilities: {x}_{y}_{z} (x = number of customers, y = number of facilities, z = ratio, e.g. 150_150_5)`
    - `indset: {x}_{y} (x = number of nodes, y = affinity, e.g. 1000_4)`
- Choose absolute folder path where generated data will live: `<root_absolute_path>`
- Choose number of solutions (we use k = 10 mostly in the paper): `<solutions_k>`
- Set environment variables:
    ```
     export MIP_NODE_SELECTION_PROBLEM=<problem>
     export MIP_NODE_SELECTION_DIFFICULTY=<difficulty>
     export MIP_NODE_SELECTION_DATA_ROOT=<root_absolute_path>
     export MIP_NODE_SELECTION_SOLUTIONS_K=<solutions_k>
    ```
- Example:
   ```
     export MIP_NODE_SELECTION_PROBLEM=setcover
     export MIP_NODE_SELECTION_DIFFICULTY=1000r_2000c_0.05d
     export MIP_NODE_SELECTION_DATA_ROOT=/opt/mip_node_selection
     export MIP_NODE_SELECTION_SOLUTIONS_K=10
   ```
- If you have trouble running any of the scripts, then try to set PYTHONPATH as the absolute path to the project root.
## Creating instances
To create instances, run 

```
python sampling/01_generate_instances.py --train_instances --valid_instances --test_instances
```

where
- `train_instances (int)` is the number of training instances, default: 300
- `valid_instances (int)` is the number of validation instances, default: 100
- `test_instances (int)` is the number of testing instances, default: 100

### Notes
- For the hard setcover instances, we did not create any training or validation
instances.
- Number of training, validation and testing instances are hard-coded.
- It may be possible to create instances that are not solvable, we suggest creating more instances than you need.

## Creating state-action pairs
To create state-action pairs, run

```
python sampling/generate_dataset.py --njobs --train_instances --valid_instances --test_instances
```

where 
- `njobs (int)` refers to the number of parallel jobs, default: 1
- `train_instances (int)` is the number of training instances to generate samples from, default: 200
- `valid_instances (int)` is the number of validation instances to generate samples from, default: 35
- `test_instances (int)` is the number of testing instances to generate samples from, default: 35

## Applying preprocessing
To apply preprocessing on state-action pairs, run:

```
python preprocessing/preprocessing.py
```

## Obtaining our ML policy
We train multiple ML models by adjusting hyperparameters and ultimately choosing the best performer. Run:

``` 
python hyperparams/hyperopt_hyperparams.py
```

Once done, check out `{MIP_NODE_SELECTION_DATA_ROOT}/hyperparams/best_trained_models/` for the best performing ML model.

## Obtaining He's policy
If you want to obtain He's policy yourself, then you'll need to set up He's work yourself. 
This involves obtaining CPLEX 12.6 and compiling it. 

Your options are:
 - Use our trained policies. You'll need to copy paste them from `he/policy/` into `{MIP_NODE_SELECTION_DATA_ROOT}/he/policy/`
 - Obtain the policies yourself (killPolicy.1 and searchPolicy.1) and paste it in:
  `{MIP_NODE_SELECTION_DATA_ROOT}/he/policy/{MIP_NODE_SELECTION_PROBLEM}_{MIP_NODE_SELECTION_DIFFICULTY}/`
 - Set `MIP_NODE_SELECTION_HE_ROOT` to point to the root of He's project, e.g.
    ```
    export MIP_NODE_SELECTION_HE_ROOT=/opt/scip-dagger-master
    ```
   and run 
   ```
   python he/train.py
   ```
   to use our script to create a policy using He's work.
### Notes 
- If you do not have He's original policy setup, then you can not do policy evaluation for that either.
- We have edited some scripts slightly of He's original setup, to make interfacing easier with our code. 
You can find the patched files in `he/patched/`.

## First solution at leaf & Optimal policy evaluation
Run
```
python evaluation/evaluate.py --transfer_mode --model_train_difficulty --resume --first_solution_only --njobs
```
where
- `transfer_mode (bool)`: Set this flag to enable transfer mode, where you typically use a model trained on easier problems to evaluate its performance on harder problems, default: False
- `model_train_difficulty (str)`: Required if transfer mode is enabled. Specify the training difficulty, so the script can find the trained policy, default: None
- `resume (bool)`: Set this flag if the script was interrupted previously and you want to continue where it was left off, default: False
- `first_solution only (bool)`: Set this flag if you only want to evaluate the performance of finding the first solution by hardcoding max number of SCIP solutions to 1. 
Note that this is not the same as stopping when the solver reaches a leaf!. Default: False
- `njobs (int)`: Number of parallel jobs, default: 1

to compare the performance on the test data of the exact policies between your best performing ML model, He's models and default SCIP models.

Check out `{MIP_NODE_SELECTION_DATA_ROOT}/evaluation/results/` for the results.


## Limited time policy evaluation
Optimal policy evaluation must have been completed before this is run.

Run
```
python evaluation/evaluate_approximation.py <optimal_policy_path> --transfer_mode --model_train_difficulty --resume --njobs
```

where
- `optimal_policy_path (str)`: is the path to the results of optimal policy evaluation, e.g.
`{MIP_NODE_SELECTION_DATA_ROOT}/evaluation/results/results_{some_timestamp}.csv` 
- `transfer_mode (bool)`: Set this flag to enable transfer mode, where you typically use a model trained on easier problems to evaluate its performance on harder problems, default: False
- `model_train_difficulty (str)`: Required if transfer mode is enabled. Specify the training difficulty, so the script can find the trained policy, default: None
- `resume (bool)`: Set this flag if the script was interrupted previously and you want to continue where it was left off, default: False
- `njobs (int)`: Number of parallel jobs, default: 1

to compare the performance on the test data of the approximate policies between your best performing ML model, He's models and default SCIP models.

Check out `{MIP_NODE_SELECTION_DATA_ROOT}/evaluation/results_approx/` for the results.