import argparse
import datetime
import glob
import subprocess
import time
import utilities
import pyscipopt as scip
import json
import torch
import os
import re
import multiprocessing as mp
import pandas as pd


from evaluation.nodepru_policy import NodepruPolicy
from evaluation.nodesel_dfs import NodeselDfs
from evaluation.nodesel_estimate import NodeselEstimate
from he.nodepru_he_policy import NodepruHePolicy
from he.nodesel_he_policy import NodeselHePolicy
from he.test import Policy
from models.neural_net import NeuralNet
from shutil import copyfile, rmtree

from evaluation.nodesel_policy_impls import NodeselPolicyRestartdfs, NodeselPolicyEstimate, NodeselPolicyScore
from sampling.nodesel_oracle import NodeselRestartdfs


def log(arr, results_file_override=None):
    # results_file_to_use = results_file
    if results_file_override is not None:
        results_file_to_use = results_file_override
    else:
        results_file_to_use = results_file
    with open(results_file_to_use, 'a') as f:
        arrstr = ','.join([f'{x}' for x in arr])
        f.write(arrstr + '\n')


def init_scip_model(instance, first_solution_only):
    m = scip.Model()
    m.setIntParam('display/verblevel', 0)
    m.readProblem(f'{instance}')
    if transfer_mode:
        utilities.init_scip_params(m, seed=seed)
    else:
        utilities.init_scip_params(m, seed=seed, heuristics=False, presolving=False)
    if first_solution_only:
        m.setIntParam('limits/solutions', 1)
    m.setIntParam('timing/clocktype', 2)
    m.setRealParam('limits/time', time_limit)
    return m


def get_optimal_objective_value_from_sol(solpath):
    try:
        with open(solpath, 'r') as f:
            line = f.readlines()[0]
        return float(re.findall(r'objective value:\s*(.*)', line)[0])
    except Exception as err:
        print(err)
        return None


def standard_solver(job_id, instance, instance_name, name, do_both, nodesel, optimal_objective_value, continue_from_last,
                    results_df,
                    init_scip_model_override=None, time_limit_override=None, custom_log=None, first_solution_only_override=None):
    policy_type = 'selection'
    if do_both:
        policy_type = 'both'

    if continue_from_last:
        is_already_done = results_df[
                              (results_df['instance'] == instance_name) &
                              (results_df['policy_type'] == policy_type) &
                              (results_df['strategy'] == name)].shape[0] > 0
        if is_already_done:
            print(
                f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Already done, skipping...')
            return

    print(f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type})...')
    start_time = time.time()

    if first_solution_only_override is not None:
        first_solution_only_used = first_solution_only_override
    else:
        first_solution_only_used = first_solution_only

    if init_scip_model_override is None:
        m = init_scip_model(instance, first_solution_only_used)
    else:
        m = init_scip_model_override(instance, time_limit_override, first_solution_only_used)

    m.includeNodesel(nodesel,
                     name='nodesel policy',
                     desc='Node selection',
                     stdpriority=999999,
                     memsavepriority=999999)
    if do_both:
        nodepru = NodepruPolicy()
        m.includeNodepru(nodepru,
                         name='nodepru policy',
                         desc='Node pruning on baseline',
                         stdpriority=999999,
                         memsavepriority=999999)
    m.optimize()
    elapsed_time = m.getSolvingTime()
    explored_nodes = m.getNNodes()
    lps_solved = m.getNLPs()
    best_obj_value_found = m.getSolObjVal(m.getBestSol())
    if optimal_objective_value is None:
        # Optimal objective value does not exist, so the ogap will be None for all problems
        ogap = None
    else:
        ogap = abs(best_obj_value_found - optimal_objective_value) / abs(optimal_objective_value)
    igap = m.getGap()
    igap = 'infinity' if m.isInfinity(igap) else igap
    nodes_pruned = 0
    if do_both:
        nodes_pruned = nodepru.number_nodes_pruned
    if explored_nodes == 0:
        prune_rate = 0
    else:
        prune_rate = nodes_pruned / explored_nodes

    print(
        f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Time to solve: {elapsed_time:.3f} s')
    print(
        f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Number of nodes explored: {explored_nodes}')
    print(
        f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Number of LPs solved: {lps_solved}')
    print(f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Optimality Gap: {ogap}')
    print(f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Integrality Gap: {igap}')
    print(f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Nodes pruned: {nodes_pruned}')
    print(f'#{job_id} | {instance_name}: Solving with nodesel: {name} ({policy_type}): Prune rate: {prune_rate:.3f}')
    res = [instance_name, policy_type, name, elapsed_time, explored_nodes, lps_solved, ogap, igap, nodes_pruned,
           prune_rate, -1, -1, best_obj_value_found]
    if custom_log is not None:
        custom_log(res)
    else:
        log(res)


def He3(job_id, instance, instance_name, strat_name, sel, pru, optimal_objective_value):
    if continue_from_last:
        is_already_done = results_df[
                              (results_df['instance'] == instance_name) &
                              (results_df['policy_type'] == strat_name) &
                              (results_df['strategy'] == 'He')].shape[0] > 0
        if is_already_done:
            print(
                f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Already done, skipping...')
            return

    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name})...')

    he_instances_dir = f'{he_eval_folder}/dat/{problem}_{difficulty}/test-{job_id}'
    if os.path.isdir(he_instances_dir):
        rmtree(he_instances_dir)
    os.makedirs(he_instances_dir)

    he_result_dir = f'{he_eval_folder}/result/{problem}_{difficulty}/test-{job_id}/{problem}_{difficulty}'

    # Copy test file to temporary folder:
    copyfile(instance, f'{he_instances_dir}/{instance_name}')
    args = ['scripts/test_bb.sh',
            '-d', f'{problem}_{difficulty}/test-{job_id}',
            '-e', f'{problem}_{difficulty}',
            '-x', '.lp',
            '-m', f'{problem}_{difficulty}']

    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    use_difficulty = difficulty
    if transfer_mode:
        use_difficulty = model_trained_difficulty
    if sel:
        args += ['-s', os.path.join(root_dir, f'he/policy/{problem}_{use_difficulty}/searchPolicy.1')]
    if pru:
        args += ['-k', os.path.join(root_dir, f'he/policy/{problem}_{use_difficulty}/killPolicy.1')]

    subprocess.call(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=he_eval_folder)
    # Remove copied file
    os.remove(f'{he_instances_dir}/{instance_name}')

    # Get solving time, number of nodes explored and number of LPs solved from results logfile
    with open(f'{he_result_dir}/{instance_name.replace(".lp", ".log")}') as f:
        log_res = f.read()

    he_solving_time = float(re.findall(r'Solving Time.*: (.*)', log_res)[0].strip())
    he_explored_nodes = int(re.findall(r'Solving Nodes.*: (.*)', log_res)[0].strip())
    he_lps_solved = int(re.findall('Number of LPs solved:\s*(.*)', log_res)[0].strip())

    primal_bound = float(re.findall(r'Primal Bound\s*:\s*(.*?)\s', log_res)[0].strip())
    dual_bound = float(re.findall(r'Dual Bound\s*:\s*(.*?)\s', log_res)[0].strip())
    if primal_bound < 0 and dual_bound >= 0 or primal_bound >= 0 and dual_bound <= 0:
        # opposite signs, so gap is 'infinity', igap from log cant be used
        igap = 'infinity'
    else:
        igap = float(re.findall(r'Gap\s*:\s*(.*?)\s', log_res)[0].strip()) / 100  # its saved in percentage
    best_obj_value_found = None
    try:
        best_obj_value_found = float(re.findall('Best solution value:\s*(.*)', log_res)[0].strip())
        if optimal_objective_value is None:
            # Optimal objective value does not exist, so the ogap will be None for all problems
            ogap = None
        else:
            ogap = abs(best_obj_value_found - optimal_objective_value) / abs(optimal_objective_value)
    except Exception as err:
        print(err)
        ogap = 'infinity'

    if pru:
        nodes_pruned = int(re.findall(r'nodes pruned\s*:\s*(.*)', log_res)[0].strip())
    else:
        nodes_pruned = 0
    prune_rate = nodes_pruned / he_explored_nodes
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Time to solve: {he_solving_time:.3f} s')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Number of nodes explored: {he_explored_nodes}')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Number of LPs solved: {he_lps_solved}')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Optimality Gap: {ogap}')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Integrality Gap: {igap}')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Nodes pruned: {nodes_pruned}')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 3.0) ({strat_name}): Prune rate: {prune_rate:.3f}')
    res = [instance_name, strat_name, 'He', he_solving_time, he_explored_nodes, he_lps_solved, ogap, igap, nodes_pruned,
           prune_rate, -1, -1, best_obj_value_found]
    log(res)


def He6(job_id, instance, instance_name, strat_name, sel, pru, optimal_objective_value, continue_from_last,
                    results_df, he_stop_first_leaf,
                    init_scip_model_override=None, time_limit_override=None, custom_log=None,
                    problem_override=None, difficulty_override=None, transfer_mode_override=None,
                    model_trained_difficulty_override=None, first_solution_only_override=None):

    if first_solution_only_override is not None:
        first_solution_only_used = first_solution_only_override
    else:
        first_solution_only_used = first_solution_only

    if pru and not sel and he_stop_first_leaf:
        # when he_stop_first_leaf is True, then we intend to only check what the solution is at the first leaf, which
        # is only really applicable to a selector and not a pruner
        return

    if problem_override is not None:
        problem_used = problem_override
    else:
        problem_used = problem

    if difficulty_override is not None:
        difficulty_used = difficulty_override
    else:
        difficulty_used = difficulty

    if transfer_mode_override is not None:
        transfer_mode_used = transfer_mode_override
    else:
        transfer_mode_used = transfer_mode

    if model_trained_difficulty_override is not None:
        model_trained_difficulty_used = model_trained_difficulty_override
    else:
        model_trained_difficulty_used = model_trained_difficulty

    if transfer_mode_used:
        difficulty_used = model_trained_difficulty_used

    if continue_from_last:
        is_already_done = results_df[
                              (results_df['instance'] == instance_name) &
                              (results_df['policy_type'] == strat_name) &
                              (results_df['strategy'] == 'He6')].shape[0] > 0
        if is_already_done:
            print(
                f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Already done, skipping...')
            return

    # He's policy (SCIP 6)...
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}))...')
    start_time = time.time()
    if init_scip_model_override is None:
        m = init_scip_model(instance, first_solution_only_used)
    else:
        m = init_scip_model_override(instance, time_limit_override, first_solution_only_used)

    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    if sel:
        nodesel_policy = NodeselHePolicy(Policy.from_file(os.path.join(root_dir, f'he/policy/{problem_used}_{difficulty_used}/searchPolicy.1')), he_stop_first_leaf)
        m.includeNodesel(nodesel=nodesel_policy,
                         name='nodesel policy',
                         desc='Node selection policy of He',
                         stdpriority=999999,
                         memsavepriority=999999)
    if pru:
        nodepru_policy = NodepruHePolicy(Policy.from_file(os.path.join(root_dir, f'he/policy/{problem_used}_{difficulty_used}/killPolicy.1')))
        m.includeNodepru(nodepru_policy,
                         name='nodepru policy',
                         desc='Node pruning policy of He',
                         stdpriority=999999,
                         memsavepriority=999999)
    m.optimize()
    # elapsed_time = time.time() - start_time
    elapsed_time = m.getSolvingTime()
    explored_nodes = m.getNNodes()
    lps_solved = m.getNLPs()
    best_obj_value_found = m.getSolObjVal(m.getBestSol())
    if optimal_objective_value is None:
        # Optimal objective value does not exist, so the ogap will be None for all problems
        ogap = None
    else:
        ogap = abs(best_obj_value_found - optimal_objective_value) / abs(optimal_objective_value)
    igap = m.getGap()
    igap = 'infinity' if m.isInfinity(igap) else igap
    nodes_pruned = 0
    if pru:
        nodes_pruned = nodepru_policy.number_nodes_pruned
    prune_rate = nodes_pruned / explored_nodes
    print(
        f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Time to solve: {elapsed_time:.3f} s')
    print(
        f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Number of nodes explored: {explored_nodes}')
    print(
        f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Number of LPs solved: {lps_solved}')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Optimality Gap: {ogap}')
    print(f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Integrality Gap: {igap}')
    print(
        f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Nodes pruned: {nodes_pruned}')
    print(
        f'#{job_id} | {instance_name}: Solving with He et al. (SCIP 6.0) ({strat_name}): Prune rate: {prune_rate:.3f}')
    strategy = 'He6'
    if he_stop_first_leaf:
        strategy += ':stop_first_leaf'
    res = [instance_name, strat_name, strategy, elapsed_time, explored_nodes, lps_solved, ogap, igap, nodes_pruned,
           prune_rate, -1, -1, best_obj_value_found]

    if he_stop_first_leaf:
        res = res[:-1]

    if custom_log is not None:
        custom_log(res)
    else:
        log(res)


def mlp_nodesel(job_id, instance, instance_name, model_name, k, on_both, on_no_children_name, NodeSelPolicyClass,
                model, config_file, optimal_objective_value):
    if continue_from_last:
        is_already_done = results_df[
                              (results_df['instance'] == instance_name) &
                              (results_df['policy_type'] == 'selection') &
                              (results_df['strategy'] == f'{model_name}_k_{k}:{on_both}:{on_no_children_name}')
                              ].shape[0] > 0
        if is_already_done:
            print(
                f'#{job_id} | {instance_name}: Solving with ML nodesel {model_name}_k_{k}:{on_both}:{on_no_children_name}: Already done, skipping...')
            return

    # 102.046 seconds -----> 47.450 seconds
    print(
        f'#{job_id} | {instance_name}: Solving with ML nodesel {model_name}_k_{k}:{on_both}:{on_no_children_name} ...')
    start_time = time.time()
    m = init_scip_model(instance, first_solution_only)
    nodesel_policy = NodeSelPolicyClass(model, config_file, seed, device, on_both,
                                        prune=False,
                                        prune_on_both=False, only_prune_steps=False,
                                        verbose=False)

    m.includeNodesel(nodesel=nodesel_policy,
                     name='nodesel policy',
                     desc='Node selection policy based on ML, uses RestartDFS on non-dive',
                     stdpriority=999999,
                     memsavepriority=999999)
    m.optimize()
    elapsed_time = time.time() - start_time
    explored_nodes = m.getNNodes()
    lps_solved = m.getNLPs()
    best_obj_value_found = m.getSolObjVal(m.getBestSol())
    if optimal_objective_value is None:
        # Optimal objective value does not exist, so the ogap will be None for all problems
        ogap = None
    else:
        ogap = abs(best_obj_value_found - optimal_objective_value) / abs(optimal_objective_value)
    igap = m.getGap()
    igap = 'infinity' if m.isInfinity(igap) else igap
    nodes_pruned = 0
    prune_rate = 0
    on_no_children_iters = nodesel_policy.on_no_children_iters
    on_children_iters = nodesel_policy.on_children_iters
    sir = 100 * on_no_children_iters / (on_no_children_iters + on_children_iters)
    oir = 100 * on_children_iters / (on_no_children_iters + on_children_iters)
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: Time to solve: {elapsed_time:.3f} s')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: Number of nodes explored: {explored_nodes}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: Number of LPs solved: {lps_solved}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: Optimality Gap: {ogap}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: Integrality Gap: {igap}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: Nodes pruned: {nodes_pruned}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: Prune rate: {prune_rate:.3f}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: No children iters: {on_no_children_iters} ({sir:.3f}%)')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name}_k_{k}:{on_both}:{on_no_children_name}: On children iters: {on_children_iters} ({oir:.3f}%)')

    res = [instance_name, 'selection', f'{model_name}_k_{k}:{on_both}:{on_no_children_name}', elapsed_time,
           explored_nodes, lps_solved, ogap, igap, nodes_pruned, prune_rate, sir, oir, best_obj_value_found]
    log(res)


def mlp_nodeselpru(job_id, instance, instance_name, model_name, k, on_both, on_no_children_name, NodeSelPolicyClass,
                   model, config_file, optimal_objective_value, prune_on_both):
    if prune_on_both:
        # no children will be seen here, so it doesnt matter which NodeSelPolicyClass is used
        # We will pick NodeselPolicyScore
        if on_no_children_name != 'Score':
            return

    if continue_from_last:
        is_already_done = results_df[
                              (results_df['instance'] == instance_name) &
                              (results_df['policy_type'] == 'both') &
                              (results_df[
                                   'strategy'] == f'{model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}')
                              ].shape[0] > 0
        if is_already_done:
            print(
                f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Already done, skipping...')
            return

    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}...')
    start_time = time.time()
    m = init_scip_model(instance, first_solution_only)
    nodesel_policy = NodeSelPolicyClass(model, config_file, seed, device, on_both,
                                        prune=True,
                                        prune_on_both=prune_on_both,
                                        only_prune_steps=False,
                                        verbose=False)

    m.includeNodesel(nodesel=nodesel_policy,
                     name='nodesel policy',
                     desc='Node selection policy based on ML, uses RestartDFS on non-dive',
                     stdpriority=999999,
                     memsavepriority=999999)

    nodepru = NodepruPolicy()
    m.includeNodepru(nodepru,
                     name='nodepru policy',
                     desc='Node pruning policy based on ML',
                     stdpriority=999999,
                     memsavepriority=999999)
    m.optimize()
    elapsed_time = time.time() - start_time
    explored_nodes = m.getNNodes()
    lps_solved = m.getNLPs()
    best_obj_value_found = m.getSolObjVal(m.getBestSol())
    if optimal_objective_value is None:
        # Optimal objective value does not exist, so the ogap will be None for all problems
        ogap = None
    else:
        ogap = abs(best_obj_value_found - optimal_objective_value) / abs(optimal_objective_value)
    igap = m.getGap()
    igap = 'infinity' if m.isInfinity(igap) else igap
    nodes_pruned = nodepru.number_nodes_pruned
    prune_rate = nodes_pruned / explored_nodes
    on_no_children_iters = nodesel_policy.on_no_children_iters
    on_children_iters = nodesel_policy.on_children_iters
    sir = 100 * on_no_children_iters / (on_no_children_iters + on_children_iters)
    oir = 100 * on_children_iters / (on_no_children_iters + on_children_iters)
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Time to solve: {elapsed_time:.3f} s')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Number of nodes explored: {explored_nodes}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Number of LPs solved: {lps_solved}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Optimality Gap: {ogap}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Integrality Gap: {igap}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Nodes pruned: {nodes_pruned}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: Prune rate: {prune_rate:.3f}')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: No children iters: {on_no_children_iters} ({sir:.3f}%)')
    print(
        f'#{job_id} | {instance_name}: Solving with ML {model_name} nodesel & nodepru {model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}: On children iters: {on_children_iters} ({oir:.3f}%)')
    res = [instance_name, 'both', f'{model_name}_k_{k}:{on_both}:{prune_on_both}:{on_no_children_name}', elapsed_time,
           explored_nodes, lps_solved, ogap, igap, nodes_pruned, prune_rate, sir, oir, best_obj_value_found]
    log(res)


def get_jobs():
    jobs_queue = mp.Queue()
    ml_models = []
    model_names = ['vanilla_1024']
    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    for model_name in model_names:
        # for k in all_ks:
        k = int(os.getenv('MIP_NODE_SELECTION_SOLUTIONS_K'))
        model_dir = os.path.join(root_dir, f'hyperparams/best_trained_models/{problem}_{model_trained_difficulty}/{model_name}_k_{k}')
        with open(os.path.join(model_dir, 'config.json'), 'r') as f:
            model_config = json.load(f)

        model = NeuralNet(model_config['num_feats'], model_config['n_hidden_layers'], model_config['hidden_units'],
                          model_config['dropout_rate'], k).to(device)
        model.load_state_dict(torch.load(f'{model_dir}/best_params.pkl'))
        ml_models.append((model_name, model, k))

    prune_on_both_arr = [True, False]
    for instance in test_instances:
        optimal_objective_value = get_optimal_objective_value_from_sol(instance + '-0.sol')
        instance_name = instance.split('/')[-1]
        # Baselines node selection and both
        nodesels = [(NodeselRestartdfs(), 'RestartDFS', False),
                    (NodeselDfs(), 'DFS', False),
                    (NodeselEstimate(), 'BestEstimate', False),
                    (NodeselRestartdfs(prune=True), 'RestartDFS', True),
                    (NodeselDfs(prune=True), 'DFS', True),
                    (NodeselEstimate(prune=True), 'BestEstimate', True)
                    ]

        if transfer_mode:
            nodesels = [nodesels[2], nodesels[5]]  # for big problems, only try NodeselEstimate selector and pruner
        for nodesel, name, do_both in nodesels:
            jobs_queue.put((standard_solver, (instance, instance_name, name, do_both, nodesel, optimal_objective_value, continue_from_last, results_df)))

        # He et al: 6 jobs total
        strats = [(True, False, 'selection'), (False, True, 'pruning'), (True, True, 'both')] # include (sel, pru), ...
        for sel, pru, strat_name in strats:
            if he_eval_folder is not None:  # cant run He3 jobs if the user has not setup He3
                jobs_queue.put((He3, (instance, instance_name, strat_name, sel, pru, optimal_objective_value)))
            jobs_queue.put((He6, (instance, instance_name, strat_name, sel, pru, optimal_objective_value, continue_from_last, results_df, False)))

        # Stop at first leaf for He:
        jobs_queue.put((He6, (instance, instance_name, 'selection', True, False, optimal_objective_value, continue_from_last, results_df, True)))


        # ML policies...: 4 models
        for model_name, model, k in ml_models:
            # Choose between the 'specificA' node selector, 'specificA AND specificB' node pruner or both
            # 'specificA' meaning 'prioChild', 'second' or 'random'
            # 'specificB' meaning 'prune_on_both' == False | True
            if transfer_mode:
                config_file = os.path.join(root_dir, f'data/samples-preprocessed/{problem}/{model_trained_difficulty}/k_{k}/train/preprocess_params.pkl')
            else:
                config_file = os.path.join(root_dir,
                                       f'data/samples-preprocessed/{problem}/{difficulty}/k_{k}/train/preprocess_params.pkl')
            if k > 1:
                # node selection                 : 3 models x 3 on_no_children x 3 on_both = 27
                # node selpru pruneonboth=false  : 3 models x 3 on_no_children x 3 on_both = 27
                # node selpru pruneonboth=true   : 3 models x 1 on_no_children x 3 on_both = 9
                on_boths = ['prioChild', 'second', 'random']
            else:
                # node selection                 : 1 model x 3 on_no_children x 1 on_both = 3
                # node selpru pruneonboth=false  : 1 model x 3_on_no_children x 1 on_both = 3
                # node selpru pruneonboth=true   : 1 model x 1 on_no_children x 1 on_both = 1
                on_boths = ['']

            on_no_children = [(NodeselPolicyRestartdfs, 'Restartdfs'),
                              (NodeselPolicyEstimate, 'Estimate'),
                              (NodeselPolicyScore, 'Score')]

            # on_no_children = [(NodeselPolicyScore, 'Score')]

            for NodeSelPolicyClass, on_no_children_name in on_no_children:

                # Only node selector: 4 models x (3 on_no_children x 3 on_both + 1) = 40
                for on_both in on_boths:
                    jobs_queue.put((mlp_nodesel, (instance, instance_name, model_name, k, on_both, on_no_children_name, NodeSelPolicyClass,
                                 model, config_file, optimal_objective_value)))

                # Both ML nodeselector & pruner: 3 models x (3 on_no_children x 3_on_both + 1) * 2 prune_on_both + 10 = 70
                for on_both in on_boths:  # TODO: REVERT THIS
                   for prune_on_both in prune_on_both_arr:
                       jobs_queue.put((mlp_nodeselpru, (instance, instance_name, model_name, k, on_both, on_no_children_name,
                                      NodeSelPolicyClass, model, config_file, optimal_objective_value, prune_on_both)))
    return jobs_queue


def run_jobs(job_id, jobs_queue, done_queue, dummy_instance, init_scip_model, first_solution_only_override=None):
    time.sleep(1)
    jobs_completed_on_process = 0
    total_elapsed_time_job = 0
    while not jobs_queue.empty():
        task, args = jobs_queue.get()
        print(f'#{job_id}: JOBS LEFT: {jobs_queue.qsize()}')
        start_time_job = time.time()
        task(job_id, *args)
        jobs_completed_on_process += 1
        elapsed_time_job = time.time() - start_time_job
        total_elapsed_time_job += elapsed_time_job
        avg_time = total_elapsed_time_job / jobs_completed_on_process
        jobs_to_go = jobs_queue.qsize()
        time_left = avg_time * jobs_to_go
        time_left_str = convert_seconds_to_time_left(time_left)
        print(f'#{job_id}: TIME LEFT: {time_left:.3f} s | {time_left_str}')

    done_queue.get()
    time.sleep(1)
    print(f'#{job_id}: DONE')
    if done_queue.empty():
        print(f'#{job_id}: DONE QUEUE EMPTY | RETURNING!')
        return

    if first_solution_only_override is not None:
        first_solution_only_used = first_solution_only_override
    else:
        first_solution_only_used = first_solution_only

    while not done_queue.empty():
        m = init_scip_model(dummy_instance, first_solution_only_used)
        m.setRealParam('limits/time', 10)
        m.optimize()


def convert_seconds_to_time_left(n):
    day = str(int(n // (24 * 3600)))

    n = n % (24 * 3600)
    hour = str(int(n // 3600))

    n %= 3600
    minutes = str(int(n // 60))

    n %= 60
    seconds = str(int(n))

    return f'{day} days, {hour.zfill(2)}:{minutes.zfill(2)}:{seconds.zfill(2)}'


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--transfer_mode',
        help='Set this flag if the model is trained on a different difficulty than the test set.',
        action='store_true'
    )

    parser.add_argument(
        '--model_train_difficulty',
        help='Required if transfer_mode is set. '
             'Usually the test set instances are more difficult than the training set instances',
        type=str,
        default=None
    )

    parser.add_argument(
        '--resume',
        help='Resume policy evaluation if interrupted earlier.',
        action='store_true'
    )

    parser.add_argument(
        '--first_solution_only',
        help='Evaluate policies by the performing of finding the first solution only.',
        action='store_true'
    )

    parser.add_argument(
        '--njobs',
        help='Number of parallel jobs.',
        type=int,
        default=1
    )

    args = parser.parse_args()

    if args.transfer_mode and args.model_train_difficulty is None:
        raise Exception('--model_train_difficulty must be set if --transfer_mode is set')

    continue_from_last = args.resume
    first_solution_only = args.first_solution_only

    required_envs = ['MIP_NODE_SELECTION_PROBLEM', 'MIP_NODE_SELECTION_DIFFICULTY', 'MIP_NODE_SELECTION_DATA_ROOT',
                     'MIP_NODE_SELECTION_SOLUTIONS_K']

    for env in required_envs:
        if env not in os.environ:
            raise Exception(f'{env} is not set')

    problem = os.getenv('MIP_NODE_SELECTION_PROBLEM')
    difficulty = os.getenv('MIP_NODE_SELECTION_DIFFICULTY')
    transfer_mode = args.transfer_mode
    if transfer_mode:
        model_trained_difficulty = args.model_train_difficulty
    else:
        model_trained_difficulty = difficulty

    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    results_dir = os.path.join(root_dir, f'evaluation/results/{problem}_{difficulty}')
    os.makedirs(results_dir, exist_ok=True)
    date_time = datetime.datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
    if continue_from_last:
        results_files = sorted(glob.glob(f'{results_dir}/results_*.csv'), key=lambda x: time.strptime(x.split('/')[-1][8:-4], "%d-%m-%Y_%H-%M-%S"))
        results_file = results_files[-1]
    else:
        results_file = f'{results_dir}/results_{date_time}.csv'

    if 'MIP_NODE_SELECTION_HE_ROOT' not in os.environ:
        print('He3 root not provided, will skip!')
        he_eval_folder = None
    else:
        he_eval_folder = os.getenv('MIP_NODE_SELECTION_HE_ROOT')

    device = torch.device('cpu')
    test_instances = sorted(glob.glob(os.path.join(root_dir, f'data/instances/{problem}/test_{difficulty}/instance_*.lp')),
                  key=utilities.filename_instance_sort_func)

    # filter instances by the existence of a solution.
    # this happens in the case when we generate sample from fewer instances than the total number of available instances
    # because only during sampling we actually solve the selected instances
    test_instances = [x for x in test_instances if os.path.exists(f'{x}-0.sol')]

    if continue_from_last:
        results_df = pd.read_csv(results_file)
    else:
        results_df = None

    print(test_instances)
    total_instance_number = len(test_instances)

    seed = 0
    time_limit = 3600
    if not continue_from_last:
        cols = ['instance', 'policy_type', 'strategy', 'solving_time', 'explored_nodes', 'lps_solved',
         'ogap', 'igap', 'nodes_pruned', 'prune_rate', 'no_children_rate', 'children_rate','best_obj_value_found']
        # if he_stop_first_leaf:
        #     cols = cols[:-1]
        log(cols)

    n_jobs = args.njobs

    jobs_queue = get_jobs()
    done_queue = mp.Queue()
    for i in range(n_jobs):
        done_queue.put(True)

    dummy_instance = test_instances[0]
    processes = []

    for i in range(n_jobs):
        p = mp.Process(target=run_jobs, args=(i,jobs_queue, done_queue, dummy_instance, init_scip_model))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()
