import argparse
import datetime
import multiprocessing as mp
import glob
import os
import time
import numpy as np

import pandas as pd
from pyscipopt import scip

import utilities
from evaluation.evaluate import get_optimal_objective_value_from_sol, standard_solver, run_jobs, log as log_orig, He6
from evaluation.nodesel_dfs import NodeselDfs
from evaluation.nodesel_estimate import NodeselEstimate
from sampling.nodesel_oracle import NodeselRestartdfs


def log(arr):
    log_orig(arr, results_file)


def init_scip_model(instance, time_limit=None, first_solution_only=False):
    m = scip.Model()
    m.setIntParam('display/verblevel', 0)
    m.readProblem(f'{instance}')
    utilities.init_scip_params(m, seed=seed, heuristics=False, presolving=False)
    m.setIntParam('timing/clocktype', 2)
    if time_limit is not None:
        m.setRealParam('limits/time', time_limit)
    if first_solution_only:
        m.setIntParam('limits/solutions', 1)
    return m


def choose_strategy(experiments_df):
    experiments_df = experiments_df[experiments_df['prune_rate'] != 'oom']
    to_floats = ['prune_rate', 'solving_time', 'explored_nodes', 'lps_solved', 'igap']
    for x in to_floats:
        experiments_df[x] = [float(y) for y in experiments_df[x].to_numpy()]
    selection = experiments_df[(experiments_df['prune_rate'] > 0) & (experiments_df['strategy'].str.startswith('vanilla_1024'))]

    grouped = selection.groupby(['policy_type', 'strategy'])
    if hard_instances:
        gap = 'igap'
        res = grouped[['solving_time', 'explored_nodes', 'lps_solved', 'igap']].agg([np.average, np.median])
    else:
        gap = 'ogap'
        res = grouped[['solving_time', 'explored_nodes', 'lps_solved', 'ogap', 'igap']].agg([np.average, np.median])
    res['score'] = res[('solving_time', 'average')] * res[(gap, 'average')]
    res = res.sort_values(['score'])
    res = res.drop('score', axis=1)
    return res.iloc[0].name[1]


def get_time_limit(instance_name):
    selected = experiments_df[
        (experiments_df['instance'] == instance_name) &
        (experiments_df['policy_type'] == 'both') & (experiments_df['strategy'] == chosen_ml_strategy)]
    if selected.shape[0] == 0:
        return None, None
    return selected['solving_time'].to_numpy(), selected.iloc[0].to_numpy()


def get_jobs():
    jobs_queue = mp.Queue()

    for instance in test_instances:
        optimal_objective_value = get_optimal_objective_value_from_sol(instance + '-0.sol')
        instance_name = instance.split('/')[-1]
        # Baselines node selection and both
        nodesels = [(NodeselRestartdfs(), 'RestartDFS', False),
                    (NodeselDfs(), 'DFS', False),
                    (NodeselEstimate(), 'BestEstimate', False),
                    ]
        time_limit, row = get_time_limit(instance_name)
        if time_limit is None:
            continue
        time_limit = float(time_limit)
        if continue_from_last:
            is_already_done = results_df[
                                  (results_df['instance'] == instance_name) &
                                  (results_df['policy_type'] == 'both') &
                                  (results_df['strategy'] == chosen_ml_strategy)
                                  ].shape[0] > 0
            if not is_already_done:
                log(row.tolist())
        else:
            log(row.tolist())
        for nodesel, name, do_both in nodesels:
            jobs_queue.put((standard_solver, (instance, instance_name, name, do_both, nodesel, optimal_objective_value,
                                              continue_from_last, results_df, init_scip_model, time_limit, log, False)))

        strats = [[True, False, 'selection'], (False, True, 'pruning'), (True, True, 'both')] # include (sel, pru), ...
        for sel, pru, strat_name in strats:
            jobs_queue.put((He6, (instance, instance_name, strat_name, sel, pru, optimal_objective_value,
                                  continue_from_last, results_df, False, init_scip_model, time_limit, log, problem,
                                  difficulty, hard_instances, model_trained_difficulty, False)))

    return jobs_queue


if __name__ == '__main__':
    required_envs = ['MIP_NODE_SELECTION_PROBLEM', 'MIP_NODE_SELECTION_DIFFICULTY', 'MIP_NODE_SELECTION_DATA_ROOT',
                     'MIP_NODE_SELECTION_SOLUTIONS_K']

    for env in required_envs:
        if env not in os.environ:
            raise Exception(f'{env} is not set')

    parser = argparse.ArgumentParser()

    parser.add_argument(
        'optimal_policy_path',
        help='Path to optimal policy results',
        type=str,
    )

    parser.add_argument(
        '--transfer_mode',
        help='Set this flag if the model is trained on a different difficulty than the test set.',
        action='store_true'
    )

    parser.add_argument(
        '--model_train_difficulty',
        help='Required if transfer_mode is set. '
             'Usually the test set instances are more difficult than the training set instances',
        type=str,
        default=None
    )

    parser.add_argument(
        '--resume',
        help='Resume policy evaluation if interrupted earlier.',
        action='store_true'
    )

    parser.add_argument(
        '--njobs',
        help='Number of parallel jobs.',
        type=int,
        default=1
    )

    args = parser.parse_args()

    continue_from_last = args.resume
    hard_instances = args.transfer_mode
    problem = os.getenv('MIP_NODE_SELECTION_PROBLEM')
    difficulty = os.getenv('MIP_NODE_SELECTION_DIFFICULTY')
    if args.transfer_mode:
        model_trained_difficulty = args.model_train_difficulty
    else:
        model_trained_difficulty = difficulty

    seed = 0

    policy_results = args.optimal_policy_path

    experiments_df = pd.read_csv(policy_results)
    chosen_ml_strategy = choose_strategy(experiments_df)
    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    test_instances = sorted(glob.glob(os.path.join(root_dir, f'data/instances/{problem}/test_{difficulty}/instance_*.lp')),
                            key=utilities.filename_instance_sort_func)

    # filter instances by the existence of a solution.
    # this happens in the case when we generate sample from fewer instances than the total number of available instances
    # because only during sampling we actually solve the selected instances
    test_instances = [x for x in test_instances if os.path.exists(f'{x}-0.sol')]

    print(test_instances)
    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    results_dir = os.path.join(root_dir, f'evaluation/results_approx/{problem}_{difficulty}')
    os.makedirs(results_dir, exist_ok=True)
    date_time = datetime.datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
    if continue_from_last:
        results_files = sorted(glob.glob(f'{results_dir}/results_*.csv'),
                               key=lambda x: time.strptime(x.split('/')[-1][8:-4], "%d-%m-%Y_%H-%M-%S"))
        results_file = results_files[-1]
    else:
        results_file = f'{results_dir}/results_{date_time}.csv'

    if continue_from_last:
        results_df = pd.read_csv(results_file)
    else:
        results_df = None

    if not continue_from_last:
        log(['instance', 'policy_type', 'strategy', 'solving_time', 'explored_nodes', 'lps_solved',
         'ogap', 'igap', 'nodes_pruned', 'prune_rate', 'no_children_rate', 'children_rate','best_obj_value_found'])

    nworkers = args.njobs
    jobs_queue = get_jobs()
    done_queue = mp.Queue()
    for i in range(nworkers):
        done_queue.put(True)

    dummy_instance = test_instances[0]
    processes = []

    for i in range(nworkers):
        p = mp.Process(target=run_jobs, args=(i,jobs_queue, done_queue, dummy_instance, init_scip_model, False))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()
