import math

from evaluation.nodesel_baseline import NodeselBaseline


class NodeselEstimate(NodeselBaseline):
    def __init__(self, prune=False):
        super().__init__(prune)
        self.minplungedepth = -1
        self.maxplungedepth = -1
        self.maxplungequot = 0.25
        self.bestnodefreq = 10
        self.breadthfirstdepth = -1
        self.plungeoffset = 0
        self.iters = 0
        self.scip_inf = None

    def nodeinit(self):
        self.scip_inf = self.model.infinity()

    def nodeselect(self, model=None):
        if model is None:
            model = self.model
        if self.scip_inf is None:
            self.scip_inf = model.infinity()
        self.iters += 1
        # check if the breadth-first search should be applied
        if model.getDepth() <= self.breadthfirstdepth:
            node = model.getPrioSibling()
            if node is not None:
                return {'selnode': node}

            node = model.getPrioChild()
            if node is not None:
                self.handle_pruning(node, model)
                return {'selnode': node}

        if self.bestnodefreq == 0:
            bestnodefreq = math.inf
        else:
            bestnodefreq = self.bestnodefreq

        # check if we dont want to do plunging yet
        if model.getNNodes() < self.plungeoffset:
            # we dont want to plunge yet: select best node from the tree
            if model.getNNodes() % bestnodefreq == 0:
                selnode = model.getBestboundNode()
            else:
                selnode = model.getBestNode()
            return {'selnode': selnode}

        # calculate minimal and maximal plunging depth
        minplungedepth = self.minplungedepth
        maxplungedepth = self.maxplungedepth
        maxplungequot = self.maxplungequot
        if minplungedepth == -1:
            minplungedepth = int(model.getMaxDepth()/10)
            if model.getNStrongbranchLPIterations() > 2*model.getNNodeLPIterations():
                minplungedepth += 10
            if maxplungedepth >= 0:
                minplungedepth = min(minplungedepth, maxplungedepth)

        if maxplungedepth == -1:
            maxplungedepth = int(model.getMaxDepth()/2)
        maxplungedepth = max(maxplungedepth, minplungedepth)

        # check, if we exceeded the maximal plunging depth
        plungedepth = model.getPlungeDepth()
        if plungedepth > maxplungedepth:
            # we dont want to plunge again: select best node from the tree
            if model.getNNodes() % bestnodefreq == 0:
                selnode = model.getBestboundNode()
            else:
                selnode = model.getBestNode()
        else:
            # get global lower and cutoff bound
            lowerbound = model.getLowerbound()
            cutoffbound = model.getCutoffbound()

            # if we didnt find a solution yet, the cutoff bound is usually very bad:
            # use only 20% of the gap as cutoff bound
            if model.getNSolsFound() == 0:
                cutoffbound = lowerbound + 0.2*(cutoffbound - lowerbound)

            # check, if plunging is forced at the current depth
            if plungedepth < minplungedepth:
                maxbound = model.infinity()
            else:
                # calculate maximal plunging bound
                maxbound = lowerbound + maxplungequot * (cutoffbound - lowerbound)

            # we want to plunge again: prefer children over siblings, and siblings over leaves,
            # but only select a child or sibling, if its estimate is small enough;
            # prefer using nodes with higher node selection priority assigned by the branching rule
            node = model.getPrioChild()
            if node is not None and node.getEstimate() < maxbound:
                selnode = node
                self.handle_pruning(selnode, model)
            else:
                node = model.getBestChild()
                if node is not None and node.getEstimate() < maxbound:
                    selnode = node
                    self.handle_pruning(selnode, model)
                else:
                    node = model.getPrioSibling()
                    if node is not None and node.getEstimate() < maxbound:
                        selnode = node
                    else:
                        node = model.getBestSibling()
                        if node is not None and node.getEstimate() < maxbound:
                            selnode = node
                        else:
                            if model.getNNodes() % bestnodefreq == 0:
                                selnode = model.getBestboundNode()
                                if selnode is None:
                                    selnode = model.getBestNode()
                            else:
                                selnode = model.getBestNode()
                                if selnode is None:
                                    selnode = model.getBestboundNode()

        if selnode is None:
            model.interruptSolve()

        return {'selnode': selnode}

    def nodecomp(self, node1, node2):
        estimate1 = node1.getEstimate()
        estimate2 = node2.getEstimate()

        ispositiveinf = estimate1 == self.scip_inf and estimate2 == self.scip_inf
        isnegativeinf = -estimate1 == self.scip_inf and -estimate2 == self.scip_inf
        iseq = estimate1 == estimate2

        if ispositiveinf or isnegativeinf or iseq:
            lowerbound1 = node1.getLowerbound()
            lowerbound2 = node2.getLowerbound()
            if lowerbound1 < lowerbound2:
                return -1
            elif lowerbound1 > lowerbound2:
                return 1
            else:
                nodetype1 = node1.getType()
                nodetype2 = node2.getType()
                nodetype_child = 3
                nodetype_sibling = 2
                if nodetype1 == nodetype_child and nodetype2 != nodetype_child:
                    return -1
                elif nodetype1 != nodetype_child and nodetype2 == nodetype_child:
                    return 1
                elif nodetype1 == nodetype_sibling and nodetype2 != nodetype_sibling:
                    return -1
                elif nodetype1 != nodetype_sibling and nodetype2 == nodetype_sibling:
                    return 1
                else:
                    depth1 = node1.getDepth()
                    depth2 = node2.getDepth()
                    if depth1 < depth2:
                        return -1
                    elif depth1 > depth2:
                        return 1
                    else:
                        return 0

        if estimate1 < estimate2:
            return -1

        assert estimate1 > estimate2
        return 1
