import time

from pyscipopt import scip
from torch import nn

import utilities
import numpy as np
import preprocessing.apply_preprocessing as apply_preprocessing
import torch


class NodeselPolicy(scip.Nodesel):
    def __init__(self, policy, config_file, seed, device, on_both,
                 prune_on_both, only_prune_steps, verbose=False):
        super().__init__()
        self.policy = policy
        self.config_file = config_file
        self.seed = seed
        self.rng = np.random.RandomState(seed)
        self.device = device
        self.verbose = verbose
        self.on_both = on_both
        self.prune_on_both = prune_on_both
        self.only_prune_steps = only_prune_steps
        self.var_name_to_index = None
        self.config_preprocessing = None
        self.state_buffer = {}

        self.extracting_state_time_sum = 0
        self.preprocessing_time_sum = 0
        self.policy_time_sum = 0
        self.total_time_sum = 0
        self.niters = 0

        self.ex_vars_sum = 0
        self.ex_node_sum = 0
        self.ex_gl_sum = 0
        self.ex_comb_sum = 0

        self.ex_vars_first_half_sum = 0
        self.ex_vars_second_half_sum = 0
        self.ex_vars_finding_ix = 0

        self.ex_vars_pseudo_branch_cands = 0
        self.ex_vars_df_time = 0
        self.ex_vars_as_type_time = 0

        self.preprocessing_dropping_cols = 0
        self.preprocessing_removing_nan = 0
        self.preprocessing_transforming = 0
        self.preprocessing_standardizing = 0

        self.total_iters = 0
        self.on_no_children_iters = 0
        self.on_children_iters = 0

        self.in_no_children = False
        self.in_negative_depth = False

    def on_no_children(self):
        raise NotImplementedError()

    def on_negative_depth(self):
        raise NotImplementedError()

    def nodeselect(self):
        self.total_iters += 1
        depth = self.model.getDepth()
        if depth < 0:
            self.in_negative_depth = True
            return self.on_negative_depth()
        self.in_negative_depth = False
        openNodes = self.model.getOpenNodesData()
        children = openNodes['children']
        if len(children) == 0:
            self.in_no_children = True
            self.on_no_children_iters += 1
            return self.on_no_children()
        else:
            self.in_no_children = False
            self.on_children_iters += 1
            nodenumber = self.model.getCurrentNode().getNumber()
            if len(children) == 1:
                print(f'Num children == 1: {nodenumber} | depth: {depth}')
                # No choice to be made here
                children[0].setPolicyPrune(False)
                return self.on_no_children()
            assert(len(children) == 2)
            children_numbers = [x.getNumber() for x in children]
            # There are more children to traverse
            # Get state
            start_time_gl = time.time()

            start_time_ex = time.time()
            variable_features, self.var_name_to_index, first_half, second_half, finding_var_ix, pseudo_branch_cands, df_time, as_type_time = utilities.extract_state(self.model, self.var_name_to_index, self.state_buffer, stats=True, use_dict=True)
            self.ex_vars_sum += time.time() - start_time_ex
            self.ex_vars_first_half_sum += first_half
            self.ex_vars_second_half_sum += second_half
            self.ex_vars_finding_ix += finding_var_ix
            self.ex_vars_pseudo_branch_cands += pseudo_branch_cands
            self.ex_vars_df_time += df_time
            self.ex_vars_as_type_time += as_type_time
            start_time = time.time()
            node_features = utilities.extract_node_features(self.model, use_dict=True)
            self.ex_node_sum += time.time() - start_time
            start_time = time.time()
            global_features = utilities.extract_global_features(self.model, use_dict=True)
            self.ex_gl_sum += time.time() - start_time
            start_time = time.time()
            combined = {**node_features, **variable_features, **global_features}
            self.ex_comb_sum += time.time() - start_time
            self.extracting_state_time_sum += time.time() - start_time_ex

            start_time = time.time()
            self.config_preprocessing, dropping_cols, removing_nan, transforming, standardizing = apply_preprocessing.one_sample_dict(combined, self.config_preprocessing, self.config_file)
            self.preprocessing_time_sum += time.time() - start_time
            self.preprocessing_dropping_cols += dropping_cols
            self.preprocessing_removing_nan += removing_nan
            self.preprocessing_transforming += transforming
            self.preprocessing_standardizing += standardizing

            start_time = time.time()
            input_tensor = torch.tensor(list(combined.values())).to(self.device).float()
            input_tensor = torch.unsqueeze(input_tensor, dim=0)
            with torch.no_grad():
                self.policy.eval()
                output = self.policy(input_tensor)
                if output.shape[1] != 1:
                    _, indices = output.topk(2)
                    node_to_select = indices[0][0].item()
                else:
                    sigm = nn.Sigmoid()
                    output_item = sigm(output).item()
                    node_to_select = 1 if output_item > 0.5 else 0

            self.policy_time_sum += time.time() - start_time
            if node_to_select == 2:
                on_both = True
                if self.on_both == 'prioChild':
                    # Select priority child set by branching rule:
                    selnode = self.model.getPrioChild()
                    selnode_number = selnode.getNumber()
                    node_to_select = children_numbers.index(selnode_number)
                elif self.on_both == 'second':
                    # Select second best:
                    node_to_select = indices[0][1].item()
                elif self.on_both == 'random':
                    # Select a random node
                    node_to_select = self.rng.choice(2)
                else:
                    raise Exception('Invalid on_both parameter')
            else:
                on_both = False

            if output.shape[1] != 1:
                children[1 - node_to_select].setPolicyScore(output[0, 1 - node_to_select])
                children[node_to_select].setPolicyScore(output[0, node_to_select])
            else:
                children[1 - node_to_select].setPolicyScore(1 - output_item)
                children[node_to_select].setPolicyScore(output_item)

            if not self.prune_on_both and on_both:
                prune = False
            else:
                prune = True

            children[1 - node_to_select].setPolicyPrune(prune)
            children[node_to_select].setPolicyPrune(False)

            self.total_time_sum += time.time() - start_time_gl
            self.niters += 1
            if self.only_prune_steps:
                return super().nodeselect()
            else:
                return {"selnode": children[node_to_select]}

    def print_performance_statistics(self):
        print(f"""\n
        Node Selection Performance Statistics:\n
        \t Extracting state: \t\t{self.extracting_state_time_sum:.3f} s | { (self.extracting_state_time_sum / self.niters):.6f} s/iter\n
        \t\t Variable state: \t\t{self.ex_vars_sum:.3f} s | { (self.ex_vars_sum / self.niters):.6f} s/iter\n
        \t\t\t First half: \t\t{self.ex_vars_first_half_sum:.3f} s | { (self.ex_vars_first_half_sum / self.niters):.6f} s/iter\n
        \t\t\t Second half: \t\t{self.ex_vars_second_half_sum:.3f} s | { (self.ex_vars_second_half_sum / self.niters):.6f} s/iter\n
        \t\t\t\t Variable index: \t\t{self.ex_vars_finding_ix:.3f} s | { (self.ex_vars_finding_ix / self.niters):.6f} s/iter\n
        \t\t\t\t Pseudo branches: \t\t{self.ex_vars_pseudo_branch_cands:.3f} s | { (self.ex_vars_pseudo_branch_cands / self.niters):.6f} s/iter\n
        \t\t\t\t Creating DF: \t\t\t{self.ex_vars_df_time:.3f} s | { (self.ex_vars_df_time / self.niters):.6f} s/iter\n
        \t\t\t\t DF as_type: \t\t\t{self.ex_vars_as_type_time:.3f} s | { (self.ex_vars_as_type_time / self.niters):.6f} s/iter\n
        \t\t Node state: \t\t\t{self.ex_node_sum:.3f} s | { (self.ex_node_sum / self.niters):.6f} s/iter\n
        \t\t Global state: \t\t\t{self.ex_gl_sum:.3f} s | { (self.ex_gl_sum / self.niters):.6f} s/iter\n
        \t\t Combining states: \t\t{self.ex_comb_sum:.3f} s | { (self.ex_comb_sum / self.niters):.6f} s/iter\n
        \t Preprocessing: \t\t{self.preprocessing_time_sum:.3f} s | { (self.preprocessing_time_sum / self.niters):.6f} s/iter\n
        \t\t Dropping cols: \t\t{self.preprocessing_dropping_cols:.3f} s | { (self.preprocessing_dropping_cols / self.niters):.6f} s/iter\n
        \t\t Removing NaN: \t\t\t{self.preprocessing_removing_nan:.3f} s | { (self.preprocessing_removing_nan / self.niters):.6f} s/iter\n
        \t\t Transforming: \t\t\t{self.preprocessing_transforming:.3f} s | { (self.preprocessing_transforming / self.niters):.6f} s/iter\n
        \t\t Standardizing: \t\t{self.preprocessing_standardizing:.3f} s | { (self.preprocessing_standardizing / self.niters):.6f} s/iter\n
        \t Evaluating policy: \t{self.policy_time_sum:.3f} s | { (self.policy_time_sum / self.niters):.6f} s/iter\n
        \t Total time: \t\t\t{self.total_time_sum:.3f} s | { (self.total_time_sum / self.niters):.6f} s/iter\n
        """)

    def nodecomp(self, node1, node2):
        score1 = node1.getPolicyScore()
        score2 = node2.getPolicyScore()
        if score1 > score2:
            return -1
        elif score1 < score2:
            return 1
        else:
            return super().nodecomp(node1, node2)