import glob
import os
import selectors
import subprocess
import sys
import time
from shutil import copy
from shutil import rmtree

import utilities
from utilities import log

if __name__ == '__main__':
    required_envs = ['MIP_NODE_SELECTION_PROBLEM', 'MIP_NODE_SELECTION_DIFFICULTY', 'MIP_NODE_SELECTION_DATA_ROOT', 'MIP_NODE_SELECTION_HE_ROOT']

    for env in required_envs:
        if env not in os.environ:
            raise Exception(f'{env} is not set')

    # USER PARAMETERS #
    working_dir = os.getenv('MIP_NODE_SELECTION_HE_ROOT')
    # working_dir = '/home/kaan/workspace/scip-dagger-master'
    problem = os.getenv('MIP_NODE_SELECTION_PROBLEM')
    difficulty = os.getenv('MIP_NODE_SELECTION_DIFFICULTY')
    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')

    # START #
    he_root_dir = os.path.join(root_dir, 'he')
    os.makedirs(he_root_dir, exist_ok=True)
    logfile = os.path.join(he_root_dir, 'log')

    log(f'He working directory: {working_dir}', logfile, verbose=True)
    log(f'Problem: {problem}', logfile, verbose=True)
    log(f'Difficulty: {difficulty}', logfile, verbose=True)


    train_instances = sorted(glob.glob(os.path.join(root_dir, f'data/instances/{problem}/train_{difficulty}/*.lp')),
           key=utilities.filename_instance_sort_func)

    valid_instances = sorted(glob.glob(os.path.join(root_dir, f'data/instances/{problem}/valid_{difficulty}/*.lp')),
           key=utilities.filename_instance_sort_func)

    # filter instances by the existence of a solution.
    # this happens in the case when we generate sample from fewer instances than the total number of available instances
    # because only during sampling we actually solve the selected instances
    train_instances = [x for x in train_instances if os.path.exists(f'{x}-0.sol')]
    valid_instances = [x for x in valid_instances if os.path.exists(f'{x}-0.sol')]


    log(f'Train instances: {len(train_instances)}', logfile, verbose=True)
    log(f'Valid instances: {len(valid_instances)}', logfile, verbose=True)

    he_data_dir = f'{working_dir}/dat/{problem}_{difficulty}'

    # Clear instances if they exist
    rmtree(he_data_dir, ignore_errors=True)

    # Copy instances
    log(f'Copying instances...', logfile, verbose=True)
    instances_all = [(train_instances, 'train'), (valid_instances, 'dev')]
    for instances, type in instances_all:
        os.makedirs(f'{he_data_dir}/{type}', exist_ok=True)
        for instance in instances:
            copy(instance, f'{working_dir}/dat/{problem}_{difficulty}/{type}')

    # Clear solutions if they exist
    he_sol_dir = f'{working_dir}/solution/{problem}_{difficulty}'
    rmtree(he_sol_dir, ignore_errors=True)

    # Copy best solutions
    log(f'Copying solutions...', logfile, verbose=True)
    solutions_all = [(train_instances, 'train'), (valid_instances, 'dev')]
    for instances, type in solutions_all:
        os.makedirs(f'{he_sol_dir}/{type}', exist_ok=True)
        for instance in instances:
            solution = f'{instance}-0.sol'
            instance_name = solution.split('/')[-1].replace('.lp-0', '')
            copy(solution, f'{working_dir}/solution/{problem}_{difficulty}/{type}/{instance_name}')

    # Run training
    log(f'Training...', logfile, verbose=True)
    args = ['scripts/train_bb.sh',
            '-d', f'{problem}_{difficulty}/train',  # train instances
            '-x', '.lp',  # suffix
            '-p', '2',  # number of passes (recommended: 2)
            '-c', '2',  # SVM penalty parameter (recommended: 2)
            '-w', '8',  # weights on positive instances (recommended: 8)
            '-e', f'{problem}_{difficulty}',  # name of experiment
            '-m', f'{problem}_{difficulty}',  # name of problem
            '-n', str(len(train_instances)),  # train a policy for every n problems (we just get all)
            '-r', '1'  # frequency, always set to 1
            ]
    cmd = ' '.join(args)
    log(cmd, logfile, verbose=True)
    process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=working_dir)
    sel = selectors.DefaultSelector()
    sel.register(process.stdout, selectors.EVENT_READ)
    sel.register(process.stderr, selectors.EVENT_READ)
    nodata = False
    while not nodata:
        for key, _ in sel.select():
            data = key.fileobj.read1().decode()
            if not data:
                nodata=True
                break
            if key.fileobj is process.stdout:
                print(data, end="")
            else:
                print('ERROR: ' + data, end="", file=sys.stderr)

    time.sleep(1)

    log(f'Copying policies...', logfile, verbose=True)
    # Copy policy here, because He 6 uses it
    policy_dir = f'{working_dir}/policy/{problem}_{difficulty}/train/{problem}_{difficulty}'
    dest_dir = os.path.join(he_root_dir, f'policy/{problem}_{difficulty}')
    os.makedirs(dest_dir, exist_ok=True)
    copy(f'{policy_dir}/searchPolicy.1', dest_dir)
    copy(f'{policy_dir}/killPolicy.1', dest_dir)

    log(f'Done!', logfile, verbose=True)
