import os
import numpy as np
import pickle
import time

from hyperopt import hp, tpe, fmin, STATUS_OK, Trials
from models.train import run_with_seeds
from models.test import test


def get_vanilla_strategy_2():
    return 'vanilla_1024', 100, {
        'batch_size': 1024,
        'use_smote': False
    }


def function(params):
    lr, n_hidden_layers, hidden_units, dropout_rate = params
    hyperparams = {
        'lr': lr,
        'n_hidden_layers': int(n_hidden_layers),
        'hidden_units': int(hidden_units),
        'dropout_rate': dropout_rate
    }
    config = {
        **hyperparams,
        **{
            'problem': problem,
            'difficulty': difficulty,
            'max_epochs': 200,
            'max_epochs_no_improvements': 30,
            'training_size': 1,
            'valid_size': 1,
            'k': k
        },
        **strategy
    }
    best_train_loss, best_valid_loss, best_valid_acc = run_with_seeds(config, hyperparams, [0], verbose=False, hyperopt_mode=True)
    global best_overall_loss
    global best_overall_acc
    if best_overall_loss is None or best_valid_loss < best_overall_loss:
        best_overall_loss = best_valid_loss
        best_overall_acc = best_valid_acc
    print(f'Best loss found so far: {best_overall_loss:.3f} | acc: {best_overall_acc:.3f}')
    return {
        'loss': best_valid_loss,
        'status': STATUS_OK,
        'eval_time': time.time()
    }


def run_trials():
    trials_step = 1  # how many additional trials to do after loading saved trials. 1 = save after iteration
    max_trials = 1  # initial max_trials. put something small to not have to wait

    try:  # try to load an already saved trials object, and increase the max
        trials = pickle.load(open(exp_name, "rb"))
        print("Found saved Trials! Loading...")
        max_trials = len(trials.trials) + trials_step
        print("Rerunning from {} trials to {} (+{}) trials".format(len(trials.trials), max_trials, trials_step))
    except Exception as err:  # create a new trials object and start searching
        print(err)
        print("Could not find saved trial, creating new one...")
        trials = Trials()

    best = fmin(fn=function, space=space, algo=tpe.suggest, max_evals=max_trials, trials=trials)

    print("Best:", best)

    # save the trials object
    with open(exp_name, "wb") as f:
        pickle.dump(trials, f)

    return max_trials


if __name__ == '__main__':
    required_envs = ['MIP_NODE_SELECTION_PROBLEM', 'MIP_NODE_SELECTION_DIFFICULTY', 'MIP_NODE_SELECTION_DATA_ROOT',
                     'MIP_NODE_SELECTION_SOLUTIONS_K']

    for env in required_envs:
        if env not in os.environ:
            raise Exception(f'{env} is not set')

    problem = os.getenv('MIP_NODE_SELECTION_PROBLEM')
    difficulty = os.getenv('MIP_NODE_SELECTION_DIFFICULTY')
    k = int(os.getenv('MIP_NODE_SELECTION_SOLUTIONS_K'))

    root_dir = os.path.join(os.getenv('MIP_NODE_SELECTION_DATA_ROOT'), 'hyperparams')

    model_name, max_trials_current_strat, strategy = get_vanilla_strategy_2()

    best_overall_loss = None
    best_overall_acc = None
    start_time = time.time()

    space = [
        hp.loguniform('lr', np.log(0.00001), np.log(0.5)),
        hp.quniform('n_hidden_layers', 1, 20, q=1),
        hp.quniform('hidden_units', 1, 50, q=1),
        hp.uniform('dropout_rate', 0, 0.5)
    ]

    exp_path = os.path.join(root_dir, f'hyperopt_configs/{problem}_{difficulty}')
    os.makedirs(exp_path, exist_ok=True)
    exp_name = os.path.join(exp_path, f'{model_name}_k_{k}.hyperopt')

    # loop indefinitely and stop whenever you like
    while True:
        start_time = time.time()
        trials_done = run_trials()
        if trials_done == max_trials_current_strat:
            break
        elapsed_time = time.time() - start_time
        print(f'Elapsed time: {elapsed_time:.3f} s')
        print(f'Trials completed: {trials_done}')

    test(problem, difficulty, model_name, k)
