import os
import torch
import json
import re
import pandas as pd
from sklearn import metrics
from glob import glob
from torch import nn
from torch.utils.data.dataloader import DataLoader
from models.neural_net import NeuralNet
from models.data import Dataset
from shutil import copy


def get_saved_model(model_dir, k, device):
    with open(f'{model_dir}/config.json', 'r') as f:
        config = json.load(f)
    model = NeuralNet(n_feats=config['num_feats'], n_hidden_layers=config['n_hidden_layers'], hidden_units=config['hidden_units'],
                      dropout_rate=config['dropout_rate'], k=k).to(device=device)
    model.load_state_dict(torch.load(f'{model_dir}/best_params.pkl'))
    return model


def get_testing_data(problem, difficulty, k, device):
    path = os.path.join(os.getenv('MIP_NODE_SELECTION_DATA_ROOT'), f'data/samples-preprocessed/{problem}/{difficulty}/k_{k}/test/samples.hdf')
    dataset = Dataset(path, device=device)
    data_loader = DataLoader(dataset, batch_size=4096, num_workers=10, pin_memory=True)
    return data_loader


def get_loss_from_log(logfile):
    with open(logfile, 'r') as f:
        lines = f.read().splitlines()
        non_empty_line = None
        for i in range(len(lines)):
            line_index = len(lines) - i - 1
            line = lines[line_index]
            if len(line.strip()) > 0:
                non_empty_line = line
                break
        if non_empty_line is None:
            raise Exception('Log file is empty!')

        loss = float(re.findall(r"loss (.*) \|", non_empty_line)[0])
        return loss


def find_best_models(model_dir, topk):
    dirs = glob(f"{model_dir}/*/")
    dirs_sorted_by_loss = sorted(dirs, key=lambda dir: get_loss_from_log(dir + 'log.txt'))
    print(dirs_sorted_by_loss)
    return dirs_sorted_by_loss[:topk]


def calc_danger_rate(true, pred):
    df = pd.concat((pd.DataFrame(true, columns=['true']), pd.DataFrame(pred, columns=['pred'])), axis=1)
    left = 0
    right = 1
    dang_miscl_left = df[(df['pred'] == left) & (df['true'] == right)]
    dang_miscl_right = df[(df['pred'] == right) & (df['true'] == left)]
    dang_miscl_amount = dang_miscl_left.shape[0] + dang_miscl_right.shape[0]
    total = df.shape[0]
    return dang_miscl_amount / total


def calc_warning_rate(true, pred, danger_rate):
    df = pd.concat((pd.DataFrame(true, columns=['true']), pd.DataFrame(pred, columns=['pred'])), axis=1)
    left = 0
    right = 1
    both = 2
    warning_miscl_left = df[(df['pred'] == left) & (df['true'] == both)]
    warning_miscl_right = df[(df['pred'] == right) & (df['true'] == both)]
    total = df.shape[0]
    warning_miscl_amount = danger_rate * total + warning_miscl_left.shape[0] + warning_miscl_right.shape[0]
    return warning_miscl_amount / total


def test(problem, difficulty, model_name, k):
    models_dir = os.path.join(os.getenv('MIP_NODE_SELECTION_DATA_ROOT'), f'hyperparams/trained_models/{problem}_{difficulty}/{model_name}_k_{k}')
    models = find_best_models(models_dir, topk=3)
    print(models)
    out_dir = os.path.join(os.getenv('MIP_NODE_SELECTION_DATA_ROOT'), f'hyperparams/best_trained_models/{problem}_{difficulty}/{model_name}_k_{k}')
    device = torch.device('cuda')
    lowest_loss_model_dir = None
    lowest_loss = None
    for model_dir in models:
        model = get_saved_model(model_dir, k, device)
        test_data = get_testing_data(problem, difficulty, k, device)
        if k == 1:
            criterion = torch.nn.BCELoss()
        else:
            criterion = torch.nn.CrossEntropyLoss()
        model.eval()
        avg_acc = 0
        avg_loss = 0
        num_samples = 0
        with torch.no_grad():
            all_predicted_labels = []
            all_targets = []
            for batch in test_data:
                batch = batch.to(device)
                input = batch[:, :-1]
                target = batch[:, -1]

                output = model(input)
                if k == 1:
                    sigm = nn.Sigmoid()
                    output = sigm(output)
                    predicted_labels = torch.squeeze((output > 0.5).float())
                    loss = criterion(output, torch.unsqueeze(target, dim=1))
                else:
                    predicted_labels = torch.argmax(output, dim=1)
                    loss = criterion(output, target.long())

                avg_loss += loss.item() * batch.shape[0]

                avg_acc += torch.sum(torch.eq(predicted_labels, target)).item()
                num_samples += batch.shape[0]

                all_predicted_labels += predicted_labels.cpu().numpy().tolist()
                all_targets += target.cpu().numpy().tolist()

            dr = calc_danger_rate(all_targets, all_predicted_labels)
            if k > 1:
                wr = calc_warning_rate(all_targets, all_predicted_labels, dr)
                print(f'DR: {dr:.3f} | WR: {wr:.3f}')
                print(metrics.classification_report(all_targets, all_predicted_labels, digits=3,
                                                    target_names=['Left', 'Right', 'Both']))
            else:
                print(f'DR: {dr:.3f}')
                print(metrics.classification_report(all_targets, all_predicted_labels, digits=3, target_names=['Left', 'Right']))

        avg_loss /= num_samples
        avg_acc /= num_samples

        print(f'{avg_loss}, {avg_acc}')

        with open(f'{model_dir}/test_results.csv', 'w') as f:
            f.write(f'loss, acc\n{avg_loss}, {avg_acc}')

        if lowest_loss is None or avg_loss < lowest_loss:
            lowest_loss = avg_loss
            lowest_loss_model_dir = model_dir

    model_files = glob(lowest_loss_model_dir + '/*')
    os.makedirs(out_dir, exist_ok=True)
    for file in model_files:
        copy(file, out_dir)

    print(f'Copied {lowest_loss_model_dir} to {out_dir}')
