import datetime
import json
import time
import pickle
from torch.utils.data import SubsetRandomSampler
from torch.utils.tensorboard import SummaryWriter
from utilities import log
import torch
import torch.nn as nn
import matplotlib.pyplot as plt
import os
import numpy as np
from torch.utils.data.dataloader import DataLoader
from models.data import Dataset
from models.neural_net import NeuralNet
from torch.optim.lr_scheduler import ExponentialLR

device = torch.device('cuda')
config = None


def load_new_data(problem, difficulty, batch_size, training_size, valid_size, use_smote, k):
    training_data_loader = None
    valid_data_loader = None
    samples_dir = os.path.join(os.getenv('MIP_NODE_SELECTION_DATA_ROOT'), f'data/samples-preprocessed/{problem}/{difficulty}/k_{k}')
    # samples_dir = f'../data/samples-preprocessed/{problem}/{difficulty}/k_{k}'
    for data_type, size in [('train', training_size), ('valid', valid_size)]:
        filename = 'samples.hdf'
        if use_smote and data_type == 'train':
            filename = 'samples_smote.hdf'

        path = os.path.join(samples_dir, data_type, filename)
        dataset = Dataset(path, device=device)
        dataset_size = len(dataset)
        indices = list(range(dataset_size))
        np.random.shuffle(indices)

        if size <= 1:
            size = int(np.floor(size * dataset_size))

        indices = indices[:size]

        # Creating PT data samplers and loaders:
        sampler = SubsetRandomSampler(indices)
        data_loader = DataLoader(dataset, batch_size=batch_size, sampler=sampler,
                                      num_workers=10,
                                      pin_memory=True)
        if data_type == 'train':
            training_data_loader = data_loader
        else:
            valid_data_loader = data_loader

    with open(os.path.join(samples_dir, 'train/preprocess_params.pkl'), 'rb') as f:
        num_feats = pickle.load(f)['num_feats']

    return training_data_loader, valid_data_loader, num_feats


def epoch_func(batches, training, optimizer, model, criterion, k):
    avg_loss = 0
    avg_acc = 0
    num_samples = 0
    for batch_index, batch in enumerate(batches):  # batch shape: (batch, feature)
        batch = batch.to(device)
        input = batch[:, :-1]
        target = batch[:, -1]
        if training:
            model.zero_grad()

        output = model(input)
        if k == 1:
            sigm = nn.Sigmoid()
            output = sigm(output)
            loss = criterion(output, torch.unsqueeze(target, dim=1))
            predicted_labels = torch.squeeze((output > 0.5).float())
        else:
            loss = criterion(output, target.long())
            predicted_labels = torch.argmax(output, dim=1)

        avg_loss += loss.item() * batch.shape[0]
        avg_acc += torch.sum(torch.eq(predicted_labels, target)).item()

        if training:
            loss.backward()  # Does back propagation and calculates gradients
            optimizer.step()  # Updates the weights accordingly

        num_samples += batch.shape[0]

    avg_loss /= num_samples
    avg_acc /= num_samples

    return avg_loss, avg_acc


def create_plots(epoch_errs, epoch_loss, save_dir=None):
    fig = plt.figure()
    plt.plot(epoch_errs)
    plt.xlabel('Epoch')
    plt.ylabel('Error')
    plt.legend(['Train', 'Valid'])
    if save_dir is not None:
        plt.savefig(save_dir + f'/errors.png')
    else:
        plt.show()
    plt.close(fig)

    fig = plt.figure()
    plt.plot(epoch_loss)
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.legend(['Train', 'Valid'])
    if save_dir is not None:
        plt.savefig(save_dir + f'/losses.png')
    else:
        plt.show()
    plt.close(fig)


def calc_baseline(data):
    zeros = 0
    ones = 0
    twos = 0
    for batch_index, batch in enumerate(data):
        batch = batch.to(device)
        labels = batch[:, -1]
        zeros += torch.sum(labels == 0).item()
        ones += torch.sum(labels == 1).item()
        twos += torch.sum(labels == 2).item()
    return max(zeros, ones, twos) / (zeros + ones + twos)


def run_with_seeds(config, hyperparams, seeds, verbose=True, hyperopt_mode=False):
    date_time = datetime.datetime.now().strftime("%d-%m-%Y_%H-%M-%S")
    exp_name = f'{"smote" if config["use_smote"] else "vanilla"}_{config["batch_size"]}_k_{config["k"]}/{date_time}'

    root_dir = os.path.join(os.getenv('MIP_NODE_SELECTION_DATA_ROOT'), 'hyperparams')
    tensorboard_path = os.path.join(root_dir, f'tensorboard/{config["problem"]}_{config["difficulty"]}/{exp_name}')
    writer = SummaryWriter(log_dir=tensorboard_path)
    writer.add_text('Training size', str(config['training_size']))
    writer.add_text('Valid size', str(config['valid_size']))
    writer.add_text('Hyperopt mode', f'{hyperopt_mode}')
    writer.add_text('Max epochs', str(config['max_epochs']))
    writer.add_text('Early stopping', str(config['max_epochs_no_improvements']))
    writer.add_text('SMOTE?', str(config['use_smote']))

    last_train_loss_avg = 0
    valid_loss_avg = 0
    valid_acc_avg = 0
    for seed in seeds:
        last_train_loss, valid_loss, valid_acc = run(config, seed, writer, exp_name, verbose, hyperopt_mode)
        last_train_loss_avg += last_train_loss
        valid_loss_avg += valid_loss
        valid_acc_avg += valid_acc

    last_train_loss_avg /= len(seeds)
    valid_loss_avg /= len(seeds)
    valid_acc_avg /= len(seeds)

    writer.add_hparams({**hyperparams, **{'training_size': config['training_size']}},
                       {'hparam/valid_loss': valid_loss_avg, 'hparam/valid_acc': valid_acc_avg})
    writer.close()
    if verbose:
        print(f'Average across {len(seeds)} seeds: Loss: {valid_loss_avg} | Acc: {valid_acc_avg}')
    return last_train_loss_avg, valid_loss_avg, valid_acc_avg


def initialise_running_dir(config, exp_name):
    root_dir = os.path.join(os.getenv('MIP_NODE_SELECTION_DATA_ROOT'), 'hyperparams')
    running_dir = os.path.join(root_dir, f'trained_models/{config["problem"]}_{config["difficulty"]}/{exp_name}')
    os.makedirs(running_dir, exist_ok=True)
    with open(f'{running_dir}/config.json', 'w') as f:
        json.dump(config, f, indent=4)
    logfile = running_dir + '/log.txt'
    return running_dir, logfile


def run(config, seed, writer, exp_name, verbose=True, hyperopt_mode=False):
    torch.manual_seed(seed)
    np.random.seed(seed)

    training_data, valid_data, num_feats = load_new_data(config['problem'], config['difficulty'], config['batch_size'], config['training_size'],
                                              config['valid_size'], config['use_smote'], config['k'])
    config['num_feats'] = num_feats
    running_dir, logfile = initialise_running_dir(config, exp_name)

    writer.add_text('running_dir', running_dir)

    log(f'Train size: {len(training_data)}', logfile, verbose)
    log(f'Valid size: {len(valid_data)}', logfile, verbose)

    model = NeuralNet(n_feats=num_feats, n_hidden_layers=config['n_hidden_layers'], hidden_units=config['hidden_units'],
                      dropout_rate=config['dropout_rate'], k=config['k']).to(device=device)

    if config['k'] == 1:
        criterion = nn.BCELoss().to(device)
    else:
        criterion = nn.CrossEntropyLoss().to(device)
    optimizer = torch.optim.Adam(model.parameters(), lr=config['lr'])

    lr_scheduler = ExponentialLR(optimizer, 0.2)
    best_valid_loss = None
    no_improvements_epochs = 0
    patience = 20
    start_patience = False
    best_valid_acc = 0
    epoch_errs = []
    epoch_loss = []
    last_train_loss = None

    baseline_train_acc = calc_baseline(training_data)
    baseline_valid_acc = calc_baseline(valid_data)
    log(f"Seed {seed} Training baseline acc: {baseline_train_acc:.3f}", logfile, verbose)
    log(f"Seed {seed} Valid baseline acc: {baseline_valid_acc:.3f}", logfile, verbose)

    total_elapsed_time = 0
    for epoch in range(config['max_epochs']):
        start_time = time.time()
        model.train()
        training_loss, training_acc = epoch_func(training_data, True, optimizer, model, criterion, config['k'])
        last_train_loss = training_loss
        log(f'Seed {seed} Epoch {epoch} train loss: {training_loss:.3f}, accuracy: {training_acc:.3f}', logfile, verbose)
        model.eval()
        with torch.no_grad():
            valid_loss, valid_acc = epoch_func(valid_data, False, None, model, criterion, config['k'])
        log(f'Seed {seed} Epoch {epoch} valid loss: {valid_loss:.3f}, accuracy: {valid_acc:.3f}', logfile, verbose)
        epoch_errs.append([1 - training_acc, 1 - valid_acc])
        epoch_loss.append([training_loss, valid_loss])

        writer.add_scalar(f'{seed}/Loss/train', training_loss, epoch)
        writer.add_scalar(f'{seed}/Loss/valid', valid_loss, epoch)
        writer.add_scalar(f'{seed}/Accuracy/train', training_acc, epoch)
        writer.add_scalar(f'{seed}/Accuracy/valid', valid_acc, epoch)
        elapsed_time = time.time() - start_time
        total_elapsed_time += elapsed_time
        log(f"Seed {seed} Epoch {epoch} elapsed time: {elapsed_time:.3f} s | total: {total_elapsed_time:.3f} s", logfile, verbose)
        if best_valid_loss is None or valid_loss < best_valid_loss:
            log(f"Seed {seed} Epoch {epoch} found best model so far, valid_loss: {valid_loss:.3f}, acc: {valid_acc:.3f}", logfile, verbose)
            best_valid_loss = valid_loss
            no_improvements_epochs = 0
            best_valid_acc = valid_acc
            start_patience = False
            torch.save(model.state_dict(), f'{running_dir}/best_params.pkl')
        elif start_patience and no_improvements_epochs == patience:
            # Early stopping
            log(f'Seed {seed} Epoch {epoch} no improvements for {patience} epochs, early stopping', logfile, verbose)
            break
        elif no_improvements_epochs == config['max_epochs_no_improvements']:
            log(f'Seed {seed} Epoch {epoch} no improvements for {config["max_epochs_no_improvements"]} epochs, lowering learning rate', logfile, verbose)
            no_improvements_epochs = 0
            lr_scheduler.step(epoch=None)
            start_patience = True
        else:
            no_improvements_epochs += 1

    log(f'Seed {seed} Best model: valid loss {best_valid_loss} | acc {best_valid_acc}', logfile, verbose)

    if hyperopt_mode:
        create_plots(epoch_errs, epoch_loss, running_dir)
    else:
        create_plots(epoch_errs, epoch_loss)

    return last_train_loss, best_valid_loss, best_valid_acc
