import glob

import pandas as pd
import math
import pickle
import os
import numpy as np

import utilities
from imblearn.over_sampling import SMOTENC


def get_instance_sequence(row_number):
    for key in row_index_to_instance_sequence.keys():
        if key >= row_number:  # since keys are sorted ascendingly
            return row_index_to_instance_sequence[key]


if __name__ == '__main__':
    required_envs = ['MIP_NODE_SELECTION_PROBLEM', 'MIP_NODE_SELECTION_DIFFICULTY', 'MIP_NODE_SELECTION_DATA_ROOT',
                     'MIP_NODE_SELECTION_SOLUTIONS_K']

    for env in required_envs:
        if env not in os.environ:
            raise Exception(f'{env} is not set')

    problem = os.getenv('MIP_NODE_SELECTION_PROBLEM')
    difficulty = os.getenv('MIP_NODE_SELECTION_DIFFICULTY')
    k = int(os.getenv('MIP_NODE_SELECTION_SOLUTIONS_K'))
    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    files = [sorted(glob.glob(os.path.join(root_dir, f'data/samples/{problem}/{difficulty}/{type_of_data}/k_{k}/instance_*.hdf')), key=utilities.filename_instance_sort_func) for type_of_data in ['train', 'valid', 'test']]

    print(f'Preprocessing, doing k={k}')
    out_dir = os.path.join(root_dir, f'data/samples-preprocessed/{problem}/{difficulty}/k_{k}')

    lengths = []
    labels = []
    row_index_to_instance_sequence = {}

    dfs = [None, None, None]
    print(f'Reading instances...')
    for type_index, type_files in enumerate(files):
        for file in type_files:
            df = pd.read_hdf(file)

            if isinstance(df, pd.DataFrame):
                big_df = dfs[type_index]
                if big_df is None:
                    big_df = df
                else:
                    big_df = pd.concat((big_df, df), axis=0)
                dfs[type_index] = big_df

    train_df, valid_df, test_df = dfs[0], dfs[1], dfs[2]

    print(f'Train shape: {train_df.shape}')
    print(f'Valid shape: {valid_df.shape}')
    print(f'Test shape: {test_df.shape}')

    removed_columns = []
    preprocess_params = {
        'removed_columns': removed_columns
    }
    removed_columns_va_te = []
    print(f'Remove features that do not change')
    for col in train_df.columns:
        value = train_df[col].values[0]
        if train_df[train_df[col] != value].shape[0] == 0:
            train_df.drop(col, axis=1, inplace=True)
            print(f'Removed column {col}')
            is_in = False
            if not is_in:
                removed_columns.append(col)

            # This is for removing it from valid_df and test_df
            removed_columns_va_te.append(col)

    # Remove the same columns from the validation and test sets as well
    for df in [valid_df, test_df]:
        for col in removed_columns_va_te:
            df.drop(col, axis=1, inplace=True)

    print(f'Fix NaN')
    for type_index, df in enumerate(dfs):
        # Fix NaN within inc_val and avg_inc_val by using -1 instead, therefore creating an extra category
        df['inc_val'] = df['inc_val'].apply(lambda x: -1 if math.isnan(x) else x)
        df['avg_inc_val'] = df['avg_inc_val'].apply(lambda x: -1 if math.isnan(x) else x)
        dfs[type_index] = df

    categorical_columns = ['left_node_is_prio', 'right_node_is_prio', 'sol_is_at_lb', 'sol_is_at_ub', 'inc_val',
                           'gap_is_infinite', 'label', 'type_0', 'type_1', 'type_2', 'type_3',
                           'basis_status_0', 'basis_status_1', 'basis_status_2', 'basis_status_3']

    print(f'Standardize non_categorical')
    standardize_params = {}
    for column in train_df.columns:
        if column in categorical_columns:
            continue
        mu = train_df[column].mean()
        std = train_df[column].std()
        train_df[column] = (train_df[column] - mu) / std

        # Use the train_df means and stds for standardizing the validation and training set
        # This is because predictions in the solver is done in real time
        valid_df[column] = (valid_df[column] - mu) / std
        test_df[column] = (test_df[column] - mu) / std

        # Save mu and std for reuse on other data
        standardize_params[column] = {
            'mu': mu,
            'std': std
        }
        dfs = [train_df, valid_df, test_df]

    preprocess_params['standardized'] = standardize_params

    print(f'Fix class imbalance using SMOTE on the training dataset')
    train_df_smote = train_df.copy(deep=True)
    categorical_cols_indices = []
    train_df_no_lab = train_df_smote.loc[:, train_df_smote.columns != 'label']
    for col_index, col in enumerate(train_df_no_lab.columns):
        if col in categorical_columns:
            categorical_cols_indices.append(col_index)

    sm = SMOTENC(categorical_features=categorical_cols_indices, random_state=33)
    train_df_X, train_df_Y = sm.fit_sample(train_df_no_lab.values,
                                           train_df_smote['label'].values)

    new_train_df = pd.DataFrame(train_df_X, columns=train_df_no_lab.columns)
    new_train_df['label'] = train_df_Y
    train_df_smote = new_train_df
    dfs.append(train_df_smote)

    print(f'SMOTE results:')
    print(f"Class Left: {np.sum(train_df_smote['label'] == 0)}")
    print(f"Class Right: {np.sum(train_df_smote['label'] == 1)}")
    print(f"Class Both: {np.sum(train_df_smote['label'] == 2)}")


    print(f'Make sure that the feature "label" is the last column')
    for type_index, df in enumerate(dfs):
        labels = df['label']
        df = df.drop('label', axis=1)
        df['label'] = labels.values
        dfs[type_index] = df

    keys = list(row_index_to_instance_sequence.keys())

    type_of_files = ['/train', '/valid', '/test', '/train']
    for type_index, type_of_file in enumerate(type_of_files):
        p = out_dir + type_of_file
        os.makedirs(p, exist_ok=True)
        if type_index == 3:
            out_file = p + '/samples_smote.hdf'
        else:
            out_file = p + '/samples.hdf'
        df = dfs[type_index]
        df.to_hdf(out_file, key='t')
        print(df.shape)

    preprocess_params['num_feats'] = df.shape[1] - 1

    print(f'Saving preprocessing data for reuse')
    with open(f'{out_dir}/train/preprocess_params.pkl', 'wb') as f:
        pickle.dump(preprocess_params, f)

    print(f'k={k}: Done')
