import os
import argparse
import multiprocessing as mp
import glob
import numpy as np
import pandas as pd

import pyscipopt as scip
import utilities
import time
from sampling.nodesel_oracle import NodeselOracle
from utilities import log


def make_samples(instances, seed, out_dir, time_limit, global_episode_count, global_amount_opt, max_samples):
    """
    Worker loop: fetch an instance, run an episode and record samples.

    Parameters
    ----------
    instances: mp.Queue
        Instance files from which to collect samples.
    seed: int
    out_dir : str
        Directory in which to write samples.
    time_limit : float in [0, 1e+20]
        Maximum running time for an episode, in seconds.
    global_episode_count: mp.Value
        Keeps track of the number of episodes, shared across workers
    global_amount_opt: mp.Value
        Keeps track of the number of samples, shared across workers
    """

    while not instances.empty():
        instance = instances.get()
        if global_amount_opt.value >= max_samples:
            log(f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, reached max, early stopping...', logfile)
            break

        log(f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, processing instance \'{instance}\'...', logfile)

        m = scip.Model()
        m.setIntParam('display/verblevel', 0)
        m.readProblem(f'{instance}')
        utilities.init_scip_params(m, seed=seed, heuristics=False, presolving=False)
        m.setIntParam('timing/clocktype', 2)
        m.setRealParam('limits/time', time_limit)
        k = int(os.getenv('MIP_NODE_SELECTION_SOLUTIONS_K'))

        solution_files = sorted(glob.glob(f'{instance}-*.sol'), key=lambda x: int(x.split('/')[-1][-5]))
        if len(solution_files) == 0:
            time.sleep(1)
            log(f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, no solution exists, solving...', logfile)
            m2 = scip.Model()
            m2.setIntParam('display/verblevel', 0)
            m2.readProblem(f'{instance}')
            m2.optimize()
            solutions = m2.getSols()
            log(f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, solved {instance}', logfile)
            if len(solutions) < k:
                log(f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, not enough solutions, '
                    f'expected {k} but was {len(solutions)}. '
                    f'Will still output sols, but discontinue this iteration', logfile)
            solutions = solutions[:min(k, len(solutions))]
            for index, solution in enumerate(solutions):
                m2.writeSol(solution, f'{instance}-{index}.sol')
            solution_files = sorted(glob.glob(f'{instance}-*.sol'), key=lambda x: int(x.split('/')[-1][-5]))
            m2.freeProb()
            m2.freeTransform()

        if k == 1 and not isinstance(solution_files, list):
            solution_files = [solution_files]

        if len(solution_files) < k:
            log(f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, '
                f'not enough solutions for {instance}: {len(solution_files)}, skipping...', logfile)
            continue

        solutions = []

        for solution_file in solution_files:
            if os.path.isfile(solution_file):
                solution = m.readSolFile(solution_file)
                solutions.append(solution)
            else:
                log(f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, {solution_file} is not a file, skipping...', logfile)
                # global_episode_count.value += 1
                continue

        is_opt_checked = {}
        is_optimal = {}

        nodesel_oracle = NodeselOracle(solutions, global_episode_count.value, instance, seed,
                 out_dir, is_opt_checked, is_optimal, global_amount_opt, max_samples)


        instance_filename = nodesel_oracle.instance_filename

        if os.path.exists(f'{out_dir}/k_all/{instance_filename}.hdf'):
            log(
                f'[w {os.getpid()}] episode {global_episode_count.value}, seed {seed}, {out_dir}/k_all/{instance_filename}.hdf already exists, skipping...',
                logfile)
            global_episode_count.value += 1
            continue

        m.includeNodesel(nodesel=nodesel_oracle,
                         name='sampler',
                         desc='Sampler for node selecting, uses RestartDFS on default',
                         stdpriority=999999,
                         memsavepriority=999999)

        m.optimize()
        m.freeProb()
        total_label_counts = sum(nodesel_oracle.k_10_label_counts)
        label_counts_norm = [f'{100*x/total_label_counts:.1f}' for x in nodesel_oracle.k_10_label_counts]
        log(f'K={k} (Left, Right, Both): {label_counts_norm}', logfile)
        log(f'{nodesel_oracle.branched_vars}', logfile)
        log(f"[w {os.getpid()}] episode {global_episode_count.value} done, "
            f"{nodesel_oracle.sample_counter} samples, {global_amount_opt.value} samples in total", logfile)
        os.rename(f'{out_dir}/k_all/{instance_filename}.hdf.tmp', f'{out_dir}/k_all/{instance_filename}.hdf')
        global_episode_count.value += 1

        # split to k different folders
        for i in range(k):
            os.makedirs(f'{out_dir}/k_{i+1}', exist_ok=True)
            df = pd.read_hdf(f'{out_dir}/k_all/{instance_filename}.hdf', key='t')
            if not isinstance(df, pd.DataFrame):
                raise Exception(f'{out_dir}/{instance_filename}.hdf is not a pandas dataframe')
            # Filter unknowns
            filtered = df[df[f'label_{i+1}'] != -1]
            non_label_cols = [x for x in df.columns if not x.startswith('label_')]
            non_label_df = filtered[non_label_cols]
            non_label_df.loc[:, 'label'] = filtered[f'label_{i+1}'].values
            non_label_df.to_hdf(f'{out_dir}/k_{i+1}/{instance_filename}.hdf', key='t')


def collect_samples(instances, out_dir, rng, n_samples, n_jobs, time_limit,
                    global_episode_count, global_amount_opt):
    """
    Runs branch-and-bound episodes on the given set of instances, and collects
    randomly (state, action) pairs from the 'vanilla-fullstrong' expert
    brancher.

    Parameters
    ----------
    instances : list
        Instance files from which to collect samples.
    out_dir : str
        Directory in which to write samples.
    rng : numpy.random.RandomState
        A random number generator for reproducibility.
    n_samples : int
        Number of samples to collect.
    n_jobs : int
        Number of jobs for parallel sampling.
    time_limit : float in [0, 1e+20]
        Maximum running time for an episode, in seconds.
    global_episode_count: mp.Value
        Keeps track of the number of episodes, shared across workers
    global_amount_opt: mp.Value
        Keeps track of the number of samples, shared across workers
    """
    os.makedirs(out_dir, exist_ok=True)

    workers = []
    instances_queue = mp.Queue()
    for instance in instances:
        instances_queue.put(instance)

    for i in range(n_jobs):
        p = mp.Process(
                target=make_samples,
                args=(instances_queue, rng.randint(2**32), out_dir, time_limit, global_episode_count,
                      global_amount_opt, n_samples),
                daemon=True)
        workers.append(p)
        p.start()

    for p in workers:
        p.join()


if __name__ == '__main__':
    required_envs = ['MIP_NODE_SELECTION_PROBLEM', 'MIP_NODE_SELECTION_DIFFICULTY', 'MIP_NODE_SELECTION_DATA_ROOT',
                     'MIP_NODE_SELECTION_SOLUTIONS_K']

    for env in required_envs:
        if env not in os.environ:
            raise Exception(f'{env} is not set')

    problem = os.getenv('MIP_NODE_SELECTION_PROBLEM')
    difficulty = os.getenv('MIP_NODE_SELECTION_DIFFICULTY')

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--train_instances',
        help='Number of training instances to generate samples from',
        type=int,
        default=200
    )
    parser.add_argument(
        '--valid_instances',
        help='Number of validation instances to generate samples from',
        type=int,
        default=35
    )

    parser.add_argument(
        '--test_instances',
        help='Number of testing instances to generate samples from',
        type=int,
        default=35
    )

    parser.add_argument(
        '-s', '--seed',
        help='Random generator seed.',
        type=utilities.valid_seed,
        default=0,
    )

    parser.add_argument(
        '-j', '--njobs',
        help='Number of parallel jobs.',
        type=int,
        default=1
    )

    args = parser.parse_args()

    max_samples = 160000000
    exploration_strategy = 'pscost'
    node_record_prob = 0.05
    time_limit = 3600

    root_dir = os.getenv('MIP_NODE_SELECTION_DATA_ROOT')
    if root_dir.endswith('/'):
        root_dir = root_dir[:-1]

    in_dirs = [os.path.join(root_dir, f'data/instances/{problem}/{type_of_data}_{difficulty}')
               for type_of_data in ['train', 'valid', 'test']]
    out_dirs = [os.path.join(root_dir, f'data/samples/{problem}/{difficulty}/{type_of_data}')
               for type_of_data in ['train', 'valid', 'test']]
    num_instances_all = [args.train_instances, args.valid_instances, args.test_instances]

    log_dir = os.path.join(root_dir, 'sampling/logs')
    os.makedirs(log_dir, exist_ok=True)

    for in_dir, out_dir, num_instances in zip(in_dirs, out_dirs, num_instances_all):
        logfile = os.path.join(log_dir, f'generate_dataset_log_{time.time()}.txt')
        log(f"In dir: {in_dir}", logfile)
        log(f"Out dir: {out_dir}", logfile)
        log(f"Seed {args.seed}", logfile)
        log(f"Njobs: {args.njobs}", logfile)
        log(f'Total CPU count: {mp.cpu_count()}', logfile)

        instances = sorted(glob.glob(f'{in_dir}/*.lp'),
                           key=utilities.filename_instance_sort_func)

        instances = instances[:num_instances]

        if max_samples == np.iinfo(np.int64).max:
            log(f"{len(instances)} instances for unlimited samples", logfile)
        else:
            log(f"{len(instances)} instances for {max_samples} samples", logfile)

        # create output directory, throws an error if it already exists
        os.makedirs(out_dir, exist_ok=True)

        rng = np.random.RandomState(args.seed)
        collect_samples(instances, out_dir, rng, max_samples,
                        args.njobs, time_limit, global_episode_count=mp.Value('i', 1),
                        global_amount_opt=mp.Value('i', 1))
