import os

import pandas as pd
import utilities
import numpy as np

from evaluation.nodesel_baseline import NodeselBaseline
from evaluation.nodesel_estimate import NodeselEstimate
from utilities import check_optimality_k_sols


class NodeselRestartdfs(NodeselBaseline):
    def __init__(self, prune=False):
        super().__init__(prune)
        self.lastrestart = 0
        self.nprocessedleaves = 0
        self.selectbestfreq = 100
        self.countonlyleaves = True

    def nodeinitsol(self):
        self.lastrestart = 0
        self.nprocessedleaves = 0

    def nodeselect(self, model=None):
        if model is None:
            model = self.model
        # decide if we want to select the node with lowest bound or the deepest node; finish the current dive in any case
        selnode = model.getPrioChild()
        if selnode is not None:
            self.handle_pruning(selnode, model)
            return {"selnode": selnode}

        # increase the number of processed leafs since we are in a leaf
        self.nprocessedleaves += 1

        nnodes = model.getNNodes()

        # check if in case of "only leaves" the number processed leaves exceeds the frequency or in the other case the
        # number of processed node does it
        if self.countonlyleaves and self.nprocessedleaves >= self.selectbestfreq \
            or not self.countonlyleaves and nnodes - self.lastrestart >= self.selectbestfreq:

            self.lastrestart = nnodes
            self.nprocessedleaves = 0
            return {"selnode": model.getBestboundNode()}
        else:
            selnode = model.getPrioSibling()
            if selnode is not None:
                return {"selnode": selnode}
            else:
                return {"selnode": model.getBestLeaf()}

    def nodecomp(self, node1, node2):
        return int(node2.getNumber() - node1.getNumber())


class NodeselOracle(NodeselEstimate):
    def __init__(self, solutions, episode, instance, seed,
                 out_dir, is_opt_checked, is_optimal, global_amount_opt, global_max_opt):
        super().__init__()
        self.episode = episode
        self.instance = instance
        self.seed = seed
        self.out_dir = out_dir
        self.is_opt_checked = is_opt_checked
        self.is_optimal = is_optimal
        self.global_amount_opt = global_amount_opt
        self.global_max_opt = global_max_opt

        self.rng = np.random.RandomState(seed)
        self.new_node = True
        self.sample_counter = 0

        self.solutions = solutions
        self.opt_specific_sols = {1: [True for _ in self.solutions]} # root node is 1, which is always optimal for every solution
        self.instance_filename = instance.split('/')[-1].split('.')[:-1][0]

        self.var_names_to_index = None
        self.state_buffer = {}
        self.k_10_label_counts = [0, 0, 0]
        self.k_1_label_counts = [0, 0, 0]
        self.branched_vars = []
        self.child_is_leaf = {}
        self.visited_count = {}

    def nodeselect(self):
        node = self.model.getCurrentNode()
        depth = self.model.getDepth()
        if depth < 0:
            return super().nodeselect()
        # try:
        #     ncands = len(self.model.getLPBranchCands()[0])
        # except Exception as err:
        #     ncands = 0
        # print(f'NUMBER OF LP BRANCH CANDS AT DEPTH {depth}: {ncands}')


        node_number = node.getNumber()
        openNodes = self.model.getOpenNodesData()
        children = openNodes['children']

        # Delete my parent if I have visited my sibling before
        if depth > 0:
            parent_number = node.getParent().getNumber()
            if parent_number in self.visited_count:
                del self.visited_count[parent_number]
                del self.opt_specific_sols[parent_number]
            else:
                self.visited_count[parent_number] = 1

        if len(children) == 0:
            # Currently at a leaf
            # Apply default node selection algorithm
            # print(f'Depth: {depth} \t|\t Node: {node_number} \t|\t Gap: {100*self.model.getGap():.3f}% \t|\t Reached leaf...')
            # best_sol = self.model.getSols()[0]
            # self.model.writeSol(best_sol, f'instance_1_debug_sol.sol')
            # self.model.interruptSolve()

            # I am a leaf, and my parent put me in self.opt_specific_sols, so get me outta here
            del self.opt_specific_sols[node_number]

            return super().nodeselect()
        else:
            if self.global_amount_opt.value >= self.global_max_opt:
                return super().nodeselect()

            # There are more children to traverse
            # Check optimality of children
            child_to_opt = [False, False]
            infos = []

            for child_index, child in enumerate(children):
                info = {}
                infos.append(info)
                child_name = 'LEFT' if child_index == 0 else 'RIGHT'
                info['parent'] = node_number
                info['depth'] = depth
                info['child_number'] = child.getNumber()
                info['child_name'] = child_name
                specific_sols = check_optimality_k_sols(self.solutions, child, self.is_opt_checked,
                                                                      self.is_optimal, info)
                self.opt_specific_sols[child.getNumber()] = []
                for sol_index, sol_is_opt in enumerate(specific_sols):
                    if not self.opt_specific_sols[node_number][sol_index]:
                        # parent is not optimal, so solution can not be optimal
                        self.opt_specific_sols[child.getNumber()].append(False)
                        info[sol_index]['optimal'] = False
                    elif sol_is_opt:
                        # parent is optimal, and the solution is for this child optimal, so this child is optimal
                        self.opt_specific_sols[child.getNumber()].append(True)
                        info[sol_index]['optimal'] = True
                    else:
                        # parent is optimal, and the solution is NOT optimal for this child, so it is not optimal
                        self.opt_specific_sols[child.getNumber()].append(False)
                        info[sol_index]['optimal'] = False

                # print(info)

                #child_is_opt = np.any(self.opt_specific_sols[child.getNumber()])

                # self.opt_specific_sols[child.getNumber()] = [
                #     sol_is_opt and self.opt_specific_sols[node_number][sol_index] for sol_index, sol_is_opt in
                #     enumerate(specific_sols)]
                # info['opt_sols'] = self.opt_specific_sols[child.getNumber()]
                # print(info)
                # print(self.opt_specific_sols[child.getNumber()])
                #child_to_opt[child_index] = child_is_opt

            # Both optimal
            # if child_to_opt[0] and child_to_opt[1]:
            #     label = 2
            # elif child_to_opt[0]:
            #     label = 0
            # elif child_to_opt[1]:
            #     label = 1
            # else:
            #     return super().nodeselect()

            variable_features, self.var_names_to_index = utilities.extract_state(self.model, self.var_names_to_index, self.state_buffer, stats=False)
            node_features = utilities.extract_node_features(self.model)
            global_features = utilities.extract_global_features(self.model)
            # parent = pd.DataFrame([{'parent': self.model.getCurrentNode().getParent().getNumber() if depth > 0 else 'is_root'}])
            combined = pd.concat([node_features, variable_features, global_features], axis=1)

            last_label = -1
            label_str = 'none'
            for k in range(1, len(self.solutions) + 1):
                child_to_opt = []
                for child_index, child in enumerate(children):
                    child_to_opt.append(np.any(self.opt_specific_sols[child.getNumber()][:k]))

                if child_to_opt[0] and child_to_opt[1]:
                    label = 2
                    label_str = 'both'
                elif child_to_opt[0]:
                    label = 0
                    label_str = 'left'
                elif child_to_opt[1]:
                    label = 1
                    label_str = 'right'
                else:
                    label = -1
                    label_str = 'none'

                combined[f'label_{k}'] = label

                last_label = label
                #
                # self.k_1_label_counts[label+1] += 1
                # if label > -1:
                #     branchvar = infos[label][0]['vars'][0]['branchvar']
                #     # print(f'Depth: {depth} \t|\t Node: {node_number} \t|\t Branched variable: {branchvar} == {label} \t|\t Next child: {children[label].getNumber()} \t|\t Gap: {100*self.model.getGap():.3f}%')
                #     self.branched_vars.append((branchvar, label))
                # else:
                #     pass
                #     # print(f'Depth: {depth} \t|\t Node: {node_number} \t|\t Label == -1 \t|\t Gap: {100*self.model.getGap():.3f}%')
                # break

            if last_label == -1:
                # Invalid node, there is no solution here at all for top-k
                return super().nodeselect()

            self.k_10_label_counts[last_label] += 1

            parent = node.getParent().getNumber() if depth > 0 else 'ROOT'
            #print(f'Node: {node_number} | Parent: {parent} | Depth: {depth} | Chose: {label_str}')

            os.makedirs(f'{self.out_dir}/k_all', exist_ok=True)
            filename = f'{self.out_dir}/k_all/{self.instance_filename}.hdf.tmp'
            combined.to_hdf(filename, key='t', format='table', append=os.path.exists(filename), index=False)

            self.global_amount_opt.value += 1
            self.sample_counter += 1

            # return {'selnode': children[last_label]}

            return super().nodeselect()