import datetime
import time

import numpy as np
import scipy.sparse as sp
import pyscipopt as scip
import pandas as pd

def log(str, logfile, verbose=True):
    str = f'[{datetime.datetime.now()}] {str}'
    if verbose:
        print(str)
    with open(logfile, mode='a') as f:
        print(str, file=f)


def init_scip_params(model, seed, heuristics=True, presolving=True, separating=True, conflict=True):

    seed = seed % 2147483648  # SCIP seed range

    # set up randomization
    model.setBoolParam('randomization/permutevars', True)
    model.setIntParam('randomization/permutationseed', seed)
    model.setIntParam('randomization/randomseedshift', seed)

    # separation only at root node
    model.setIntParam('separating/maxrounds', 0)

    # no restart
    model.setIntParam('presolving/maxrestarts', 0)

    # if asked, disable presolving
    if not presolving:
        model.setIntParam('presolving/maxrounds', 0)
        model.setIntParam('presolving/maxrestarts', 0)

    # if asked, disable separating (cuts)
    if not separating:
        model.setIntParam('separating/maxroundsroot', 0)

    # if asked, disable conflict analysis (more cuts)
    if not conflict:
        model.setBoolParam('conflict/enable', False)

    # if asked, disable primal heuristics
    if not heuristics:
        model.setHeuristics(scip.SCIP_PARAMSETTING.OFF)


def extract_state(model, var_name_to_index, buffer=None, stats=False, use_dict=False):
    """
    Compute a bipartite graph representation of the solver. In this
    representation, the variables and constraints of the MILP are the
    left- and right-hand side nodes, and an edge links two nodes iff the
    variable is involved in the constraint. Both the nodes and edges carry
    features.

    Parameters
    ----------
    model : pyscipopt.scip.Model
        The current model.
    buffer : dict
        A buffer to avoid re-extracting redundant information from the solver
        each time.
    Returns
    -------
    variable_features : dictionary of type {'names': list, 'values': np.ndarray}
        The features associated with the variable nodes in the bipartite graph.
    edge_features : dictionary of type ('names': list, 'indices': np.ndarray, 'values': np.ndarray}
        The features associated with the edges in the bipartite graph.
        This is given as a sparse matrix in COO format.
    constraint_features : dictionary of type {'names': list, 'values': np.ndarray}
        The features associated with the constraint nodes in the bipartite graph.
    """
    # buffer = None
    start_time = time.time()
    if buffer is None or model.getNNodes() == 1:
        buffer = {}

    # update state from buffer if any
    s = model.getState(buffer['scip_state'] if 'scip_state' in buffer else None)
    buffer['scip_state'] = s

    if 'state' in buffer:
        obj_norm = buffer['state']['obj_norm']
    else:
        obj_norm = np.linalg.norm(s['col']['coefs'])
        obj_norm = 1 if obj_norm <= 0 else obj_norm

    row_norms = s['row']['norms']
    row_norms[row_norms == 0] = 1

    # Column features
    n_cols = len(s['col']['types'])

    if 'state' in buffer:
        col_feats = buffer['state']['col_feats']
    else:
        col_feats = {}
        col_feats['type'] = np.zeros((n_cols, 4))  # BINARY INTEGER IMPLINT CONTINUOUS
        col_feats['type'][np.arange(n_cols), s['col']['types']] = 1
        col_feats['coef_normalized'] = s['col']['coefs'].reshape(-1, 1) / obj_norm

    col_feats['has_lb'] = ~np.isnan(s['col']['lbs']).reshape(-1, 1)
    col_feats['has_ub'] = ~np.isnan(s['col']['ubs']).reshape(-1, 1)
    col_feats['sol_is_at_lb'] = s['col']['sol_is_at_lb'].reshape(-1, 1)
    col_feats['sol_is_at_ub'] = s['col']['sol_is_at_ub'].reshape(-1, 1)
    col_feats['sol_frac'] = s['col']['solfracs'].reshape(-1, 1)
    col_feats['sol_frac'][s['col']['types'] == 3] = 0  # continuous have no fractionality
    col_feats['basis_status'] = np.zeros((n_cols, 4))  # LOWER BASIC UPPER ZERO
    col_feats['basis_status'][np.arange(n_cols), s['col']['basestats']] = 1
    col_feats['reduced_cost'] = s['col']['redcosts'].reshape(-1, 1) / obj_norm
    col_feats['age'] = s['col']['ages'].reshape(-1, 1) / (s['stats']['nlps'] + 5)
    col_feats['sol_val'] = s['col']['solvals'].reshape(-1, 1)
    col_feats['inc_val'] = s['col']['incvals'].reshape(-1, 1)
    col_feats['avg_inc_val'] = s['col']['avgincvals'].reshape(-1, 1)

    col_feat_names = [[k, ] if v.shape[1] == 1 else [f'{k}_{i}' for i in range(v.shape[1])] for k, v in col_feats.items()]
    col_feat_names = [n for names in col_feat_names for n in names]
    col_feat_vals = np.concatenate(list(col_feats.values()), axis=-1)

    variable_features = {
        'names': col_feat_names,
        'values': col_feat_vals,}

    # Row features

    if 'state' in buffer:
        row_feats = buffer['state']['row_feats']
        has_lhs = buffer['state']['has_lhs']
        has_rhs = buffer['state']['has_rhs']
    else:
        row_feats = {}
        has_lhs = np.nonzero(~np.isnan(s['row']['lhss']))[0]
        has_rhs = np.nonzero(~np.isnan(s['row']['rhss']))[0]
        row_feats['obj_cosine_similarity'] = np.concatenate((
            -s['row']['objcossims'][has_lhs],
            +s['row']['objcossims'][has_rhs])).reshape(-1, 1)
        row_feats['bias'] = np.concatenate((
            -(s['row']['lhss'] / row_norms)[has_lhs],
            +(s['row']['rhss'] / row_norms)[has_rhs])).reshape(-1, 1)

    row_feats['is_tight'] = np.concatenate((
        s['row']['is_at_lhs'][has_lhs],
        s['row']['is_at_rhs'][has_rhs])).reshape(-1, 1)

    row_feats['age'] = np.concatenate((
        s['row']['ages'][has_lhs],
        s['row']['ages'][has_rhs])).reshape(-1, 1) / (s['stats']['nlps'] + 5)

    # # redundant with is_tight
    # tmp = s['row']['basestats']  # LOWER BASIC UPPER ZERO
    # tmp[s['row']['lhss'] == s['row']['rhss']] = 4  # LOWER == UPPER for equality constraints
    # tmp_l = tmp[has_lhs]
    # tmp_l[tmp_l == 2] = 1  # LHS UPPER -> BASIC
    # tmp_l[tmp_l == 4] = 2  # EQU UPPER -> UPPER
    # tmp_l[tmp_l == 0] = 2  # LHS LOWER -> UPPER
    # tmp_r = tmp[has_rhs]
    # tmp_r[tmp_r == 0] = 1  # RHS LOWER -> BASIC
    # tmp_r[tmp_r == 4] = 2  # EQU LOWER -> UPPER
    # tmp = np.concatenate((tmp_l, tmp_r)) - 1  # BASIC UPPER ZERO
    # row_feats['basis_status'] = np.zeros((len(has_lhs) + len(has_rhs), 3))
    # row_feats['basis_status'][np.arange(len(has_lhs) + len(has_rhs)), tmp] = 1

    tmp = s['row']['dualsols'] / (row_norms * obj_norm)
    row_feats['dualsol_val_normalized'] = np.concatenate((
            -tmp[has_lhs],
            +tmp[has_rhs])).reshape(-1, 1)

    row_feat_names = [[k, ] if v.shape[1] == 1 else [f'{k}_{i}' for i in range(v.shape[1])] for k, v in row_feats.items()]
    row_feat_names = [n for names in row_feat_names for n in names]
    row_feat_vals = np.concatenate(list(row_feats.values()), axis=-1)

    constraint_features = {
        'names': row_feat_names,
        'values': row_feat_vals,}

    # Edge features
    if 'state' in buffer:
        edge_row_idxs = buffer['state']['edge_row_idxs']
        edge_col_idxs = buffer['state']['edge_col_idxs']
        edge_feats = buffer['state']['edge_feats']
    else:
        coef_matrix = sp.csr_matrix(
            (s['nzrcoef']['vals'] / row_norms[s['nzrcoef']['rowidxs']],
            (s['nzrcoef']['rowidxs'], s['nzrcoef']['colidxs'])),
            shape=(len(s['row']['nnzrs']), len(s['col']['types'])))
        coef_matrix = sp.vstack((
            -coef_matrix[has_lhs, :],
            coef_matrix[has_rhs, :])).tocoo(copy=False)

        edge_row_idxs, edge_col_idxs = coef_matrix.row, coef_matrix.col
        edge_feats = {}

        edge_feats['coef_normalized'] = coef_matrix.data.reshape(-1, 1)

    edge_feat_names = [[k, ] if v.shape[1] == 1 else [f'{k}_{i}' for i in range(v.shape[1])] for k, v in edge_feats.items()]
    edge_feat_names = [n for names in edge_feat_names for n in names]
    edge_feat_indices = np.vstack([edge_row_idxs, edge_col_idxs])
    edge_feat_vals = np.concatenate(list(edge_feats.values()), axis=-1)

    edge_features = {
        'names': edge_feat_names,
        'indices': edge_feat_indices,
        'values': edge_feat_vals,}

    if 'state' not in buffer:
        buffer['state'] = {
            'obj_norm': obj_norm,
            'col_feats': col_feats,
            'row_feats': row_feats,
            'has_lhs': has_lhs,
            'has_rhs': has_rhs,
            'edge_row_idxs': edge_row_idxs,
            'edge_col_idxs': edge_col_idxs,
            'edge_feats': edge_feats,
        }

    first_half_total = time.time() - start_time
    start_time_second_half = time.time()

    # SELECT variable, constraints
    # WHERE variable == branched_variable
    # AND constraints == constraints_of(variable)

    start_time = time.time()
    cands, *_ = model.getPseudoBranchCands()
    pseudo_branch_cands = time.time() - start_time
    best_cand = cands[0]

    start_time_var_ix = time.time()

    if var_name_to_index is None:
        index_to_var_name = [x.getVar().name for x in model.getLPColsData()]
        var_name_to_index = {}
        for i in range(len(index_to_var_name)):
            var_name_to_index[index_to_var_name[i]] = i

    variable_index = var_name_to_index[best_cand.name]  # 15.729 seconds
    finding_var_index = time.time() - start_time_var_ix

    variable_features['values'] = variable_features['values'][variable_index]
    #constraint_indices = edge_features['indices'][0, np.where(edge_features['indices'][1,:] == variable_index)].flatten()
    #constraint_features['values'] = constraint_features['values'][constraint_indices]
    # skip constraints for now, as theyre variable

    start_time = time.time()
    if use_dict:
        l = len(variable_features['names'])
        df = {}
        cols = variable_features['names']
        vals = variable_features['values']
        for i in range(l):
            df[cols[i]] = vals[i]
    else:
        # variable_features_dict = {variable_features['names'][i]: [variable_features['values'][i]] for i in range(len(variable_features['names']))}
        # df = pd.DataFrame.from_dict(variable_features_dict)
        # df = pd.DataFrame.from_records(variable_features['values'].reshape(1, len(variable_features['names'])), columns=variable_features['names'])
        #df = pd.DataFrame.from_records(variable_features['values'], columns=variable_features['names'])
        df = pd.DataFrame([variable_features['values']], columns=variable_features['names'])
    df_time = time.time() - start_time


    start_time = time.time()
    # df = df.astype({
    #     'type_0': 'int8',
    #     'type_1': 'int8',
    #     'type_2': 'int8',
    #     'type_3': 'int8',
    #     'coef_normalized': 'float64',
    #     'has_lb': 'int8',
    #     'has_ub': 'int8',
    #     'sol_is_at_lb': 'int8',
    #     'sol_is_at_ub': 'int8',
    #     'sol_frac': 'float64',
    #     'basis_status_0': 'int8',
    #     'basis_status_1': 'int8',
    #     'basis_status_2': 'int8',
    #     'basis_status_3': 'int8',
    #     'reduced_cost': 'float64',
    #     'age': 'float64',
    #     'sol_val': 'float64',
    #     'inc_val': 'float64',
    #     'avg_inc_val': 'float64'
    #
    # })
    as_type_time = time.time() - start_time
    second_half_total = time.time() - start_time_second_half
    if stats:
        return df, var_name_to_index, first_half_total, second_half_total, finding_var_index, pseudo_branch_cands, df_time, as_type_time
    else:
        return df, var_name_to_index
    # return constraint_features, variable_features


def extract_node_features(model, use_dict=False):
    # Assuming that the current node has two children
    openNodes = model.getOpenNodesData()
    children = openNodes['children']
    assert len(children) == 2
    features = []
    prio_node = model.getPrioChild()
    if prio_node is None:
        prio_node_number = None
    else:
        prio_node_number = prio_node.getNumber()
    for child in children:
        features.append(child.getLowerbound()) # is the lower (dual) bound of subtree
        features.append(child.getEstimate())
        features.append(child.getBranchBound())
        if prio_node_number is None:
            features.append(0)
        else:
            if child.getNumber() == prio_node_number:
                features.append(1)
            else:
                features.append(0)
    columns = ['left_node_lb', 'left_node_estimate', 'left_node_branch_bound', 'left_node_is_prio', 'right_node_lb',
                                      'right_node_estimate', 'right_node_branch_bound', 'right_node_is_prio']
    if use_dict:
        df = {}
        for i, c in enumerate(columns):
            df[c] = features[i]
        return df
    else:
        return pd.DataFrame([features], columns=columns)
    # return pd.DataFrame([features], columns=['left_node_lb', 'left_node_estimate', 'left_node_branch_bound', 'right_node_lb', 'right_node_estimate', 'right_node_branch_bound']).astype({
    #     'left_node_lb': 'float64',
    #     'left_node_estimate': 'float64',
    #     'left_node_branch_bound': 'float64',
    #     'right_node_lb': 'float64',
    #     'right_node_estimate': 'float64',
    #     'right_node_branch_bound': 'float64'
    # })


def extract_global_features(model, use_dict=False):
    global_feats = model.getGlobalFeatures()
    global_feats['gap_is_infinite'] = int(global_feats['gap_is_infinite'])
    global_feats['depth'] = model.getDepth()
    global_feats['n_strongbranch_lp_iterations'] = model.getNStrongbranchLPIterations()
    global_feats['n_node_lp_iterations'] = model.getNNodeLPIterations()
    global_feats['max_depth'] = model.getMaxDepth()

    del global_feats['number_of_sols_found']
    if use_dict:
        return global_feats
    else:
        return pd.DataFrame([global_feats])
    # return pd.DataFrame([global_feats]).astype({
    #     'global_upper_bound': 'float64',
    #     'global_lower_bound': 'float64',
    #     'integrality_gap': 'float64',
    #     'gap_is_infinite': 'int8',
    #     'depth': 'int64'
    # })

def valid_seed(seed):
    """Check whether seed is a valid random seed or not."""
    seed = int(seed)
    if seed < 0 or seed > 2**32 - 1:
        raise argparse.ArgumentTypeError(
                "seed must be any integer between 0 and 2**32 - 1 inclusive")
    return seed


def compute_extended_variable_features(state, candidates):
    """
    Utility to extract variable features only from a bipartite state representation.

    Parameters
    ----------
    state : dict
        A bipartite state representation.
    candidates: list of ints
        List of candidate variables for which to compute features (given as indexes).

    Returns
    -------
    variable_states : np.array
        The resulting variable states.
    """
    constraint_features, edge_features, variable_features = state
    constraint_features = constraint_features['values']
    edge_indices = edge_features['indices']
    edge_features = edge_features['values']
    variable_features = variable_features['values']

    cand_states = np.zeros((
        len(candidates),
        variable_features.shape[1] + 3*(edge_features.shape[1] + constraint_features.shape[1]),
    ))

    # re-order edges according to variable index
    edge_ordering = edge_indices[1].argsort()
    edge_indices = edge_indices[:, edge_ordering]
    edge_features = edge_features[edge_ordering]

    # gather (ordered) neighbourhood features
    nbr_feats = np.concatenate([
        edge_features,
        constraint_features[edge_indices[0]]
    ], axis=1)

    # split neighborhood features by variable, along with the corresponding variable
    var_cuts = np.diff(edge_indices[1]).nonzero()[0]+1
    nbr_feats = np.split(nbr_feats, var_cuts)
    nbr_vars = np.split(edge_indices[1], var_cuts)
    assert all([all(vs[0] == vs) for vs in nbr_vars])
    nbr_vars = [vs[0] for vs in nbr_vars]

    # process candidate variable neighborhoods only
    for var, nbr_id, cand_id in zip(*np.intersect1d(nbr_vars, candidates, return_indices=True)):
        cand_states[cand_id, :] = np.concatenate([
            variable_features[var, :],
            nbr_feats[nbr_id].min(axis=0),
            nbr_feats[nbr_id].mean(axis=0),
            nbr_feats[nbr_id].max(axis=0)])

    cand_states[np.isnan(cand_states)] = 0

    return cand_states



def check_optimality_pre(node, is_opt_checked, is_optimal, node_number):
    if is_opt_checked.get(node_number, False):
        return is_optimal[node_number]
    if node.getDepth() == 0:
        is_opt_checked[node_number] = True
        is_optimal[node_number] = True
        return True
    if node.getDepth() == 1:
        # parent is root, which is always optimal
        return None
    parent_number = node.getParent().getNumber()
    if not is_opt_checked.get(parent_number, False):
        raise Exception('Parent is not opt checked')
    if not is_optimal[parent_number]:
        is_opt_checked[node_number] = True
        is_optimal[node_number] = False
        return False
    return None


def check_optimality(solution, node, is_opt_checked, is_optimal, info=None):
    if info is None:
        info = {}
    node_number = node.getNumber()
    # if is_opt_checked.get(node_number, False):
    #     return is_optimal[node_number]
    # if node.getDepth() == 0:
    #     is_opt_checked[node_number] = True
    #     is_optimal[node_number] = True
    #     return True
    # parent_number = node.getParent().getNumber()
    # if not is_opt_checked.get(parent_number, False):
    #     raise Exception('Parent is not opt checked')
    # if not is_optimal[node.getParent().getNumber()]:
    #     is_opt_checked[node_number] = True
    #     is_optimal[node_number] = False
    #     return False
    pre = check_optimality_pre(node, is_opt_checked, is_optimal, node_number)
    if pre is not None:
        info['opt_by_precondition'] = pre
        return pre
    node.checkOptimal(solution, info)
    is_opt_checked[node_number] = True
    node_is_opt = node.isOptimal()
    is_optimal[node_number] = node_is_opt
    return node_is_opt


def check_optimality_k_sols(solutions, node, is_opt_checked, is_optimal, info=None):
    if info is None:
        info = {}
    node_number = node.getNumber()
    #pre = check_optimality_pre(node, is_opt_checked, is_optimal, node_number)
    specific_sols_optimal = []
    #if pre is False:
    #    info['opt_by_precondition'] = pre
    #    return pre, [False for _ in solutions]
    is_opt_for_one_sol = False
    for i, solution in enumerate(solutions):
        inf = {}
        info[i] = inf
        node.checkOptimal(solution, inf)
        node_is_opt = node.isOptimal()
        specific_sols_optimal.append(node_is_opt)
        #is_opt_for_one_sol = is_opt_for_one_sol or node_is_opt
    is_opt_checked[node_number] = True
    #is_optimal[node_number] = is_opt_for_one_sol
    return specific_sols_optimal


def filename_sort_func(a):
    x = a.split('/')[-1].split('.')[0].split('_')[1:]
    # 1_0: 01
    # 2_0: 02
    # 3_0: 03
    # 9_0: 09

    # 1_1: 11
    # 2_1: 12

    # 1_1: 1*1 + 1 = 3
    # 2_1
    # a_b: b*a + a = (b+1)*a
    a = int(x[0])
    b = int(x[1])

    return a*10 + b

def filename_instance_sort_func(a):
    return int(a.split('/')[-1].split('.')[0].split('_')[-1])

#    return (b+1)*a